#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dynet as dy
import numpy as np
import sys
import copy
import random

from cube.generic_networks.character_embeddings import CharacterNetwork
from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder


class BDRNNTagger:
    def __init__(self, tagger_config, encodings, aux_softmax_weight=0.2, runtime=False, num_languages=1):
        self.config = tagger_config
        self.encodings = encodings

        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model)  # dy.MomentumSGDTrainer(self.model)
        self.trainer.set_sparse_updates(False)
        self.character_network = CharacterNetwork(100, encodings, rnn_size=200, rnn_layers=2,
                                                  embeddings_size=self.config.input_size,
                                                  model=self.model, runtime=runtime,
                                                  lang_embeddings_size=self.config.language_embedding_size)

        self.unknown_word_embedding = self.model.add_lookup_parameters((1, self.config.input_size))
        self.holistic_word_embedding = self.model.add_lookup_parameters(
            (len(encodings.word2int), self.config.input_size))
        self.language_embeddings = self.model.add_lookup_parameters(
            (num_languages, self.config.language_embedding_size))

        self.char_proj_w = self.model.add_parameters((self.config.input_size, self.config.input_size))
        self.hol_proj_w = self.model.add_parameters((self.config.input_size, self.config.input_size))
        self.lang_proj_w = self.model.add_parameters((self.config.input_size, self.config.language_embedding_size))

        self.bdrnn_fw = []
        self.bdrnn_bw = []
        rnn_input_size = self.config.input_size  # self.embeddings.word_embeddings_size

        aux_softmax_input_size = 0
        index = 0
        for layer_size in self.config.layers:
            if runtime:
                self.bdrnn_fw.append(dy.VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
                self.bdrnn_bw.append(dy.VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
            else:
                self.bdrnn_fw.append(orthonormal_VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
                self.bdrnn_bw.append(orthonormal_VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
            rnn_input_size = layer_size * 2 + self.config.language_embedding_size
            index += 1
            if index == self.config.aux_softmax_layer:
                aux_softmax_input_size = rnn_input_size

        self.mlps = []
        for _ in range(3):  # upos, xpos and attrs
            mlp_w = []
            mlp_b = []
            input_sz = self.config.layers[-1] * 2 + self.config.language_embedding_size
            for l_size in self.config.presoftmax_mlp_layers:
                mlp_w.append(self.model.add_parameters((l_size, input_sz)))
                mlp_b.append(self.model.add_parameters((l_size)))
                input_sz = l_size
            self.mlps.append([mlp_w, mlp_b])

        softmax_input_size = self.config.presoftmax_mlp_layers[-1]
        self.softmax_upos_w = self.model.add_parameters((len(self.encodings.upos2int), softmax_input_size))
        self.softmax_upos_b = self.model.add_parameters((len(self.encodings.upos2int)))
        self.softmax_xpos_w = self.model.add_parameters((len(self.encodings.xpos2int), softmax_input_size))
        self.softmax_xpos_b = self.model.add_parameters((len(self.encodings.xpos2int)))
        self.softmax_attrs_w = self.model.add_parameters((len(self.encodings.attrs2int), softmax_input_size))
        self.softmax_attrs_b = self.model.add_parameters((len(self.encodings.attrs2int)))

        self.aux_softmax_upos_w = self.model.add_parameters((len(self.encodings.upos2int), aux_softmax_input_size))
        self.aux_softmax_upos_b = self.model.add_parameters((len(self.encodings.upos2int)))
        self.aux_softmax_xpos_w = self.model.add_parameters((len(self.encodings.xpos2int), aux_softmax_input_size))
        self.aux_softmax_xpos_b = self.model.add_parameters((len(self.encodings.xpos2int)))
        self.aux_softmax_attrs_w = self.model.add_parameters((len(self.encodings.attrs2int), aux_softmax_input_size))
        self.aux_softmax_attrs_b = self.model.add_parameters((len(self.encodings.attrs2int)))

        self.aux_softmax_weight = aux_softmax_weight
        self.losses = []

    def tag(self, seq, lang_id=0):
        dy.renew_cg()
        softmax_list, aux_softmax_list = self._predict(seq, lang_id=lang_id)
        label_list = []
        for softmax in softmax_list:
            label_list.append([self.encodings.upos_list[np.argmax(softmax[0].npvalue())],
                               self.encodings.xpos_list[np.argmax(softmax[1].npvalue())],
                               self.encodings.attrs_list[np.argmax(softmax[2].npvalue())]])
        return label_list

    def learn(self, seq, lang_id=0):
        # dy.renew_cg()
        softmax_list, aux_softmax_list = self._predict(seq, lang_id=lang_id, runtime=False)
        losses = []
        for entry, softmax, aux_softmax in zip(seq, softmax_list, aux_softmax_list):
            upos_index = self.encodings.upos2int[entry.upos]
            xpos_index = self.encodings.xpos2int[entry.xpos]
            attrs_index = self.encodings.attrs2int[entry.attrs]

            losses.append(-dy.log(dy.pick(softmax[0], upos_index)))
            losses.append(-dy.log(dy.pick(softmax[1], xpos_index)))
            losses.append(-dy.log(dy.pick(softmax[2], attrs_index)))
            losses.append(-dy.log(dy.pick(aux_softmax[0], upos_index)) * (self.aux_softmax_weight / 3))
            losses.append(-dy.log(dy.pick(aux_softmax[1], xpos_index)) * (self.aux_softmax_weight / 3))
            losses.append(-dy.log(dy.pick(aux_softmax[2], attrs_index)) * (self.aux_softmax_weight / 3))

        # loss = dy.average(losses)
        # loss_val = loss.value()
        # loss.backward()
        # self.trainer.update()
        # return loss_val
        self.losses.append(dy.esum(losses))

    def start_batch(self):
        self.losses = []
        dy.renew_cg()

    def end_batch(self):
        total_loss_val = 0
        if len(self.losses) > 0:
            total_loss = dy.esum(self.losses)
            self.losses = []
            total_loss_val = total_loss.value()
            total_loss.backward()
            self.trainer.update()
        return total_loss_val

    def _predict(self, seq, runtime=True, lang_id=0):
        softmax_list = []
        aux_softmax_list = []
        x_list = []
        lang_emb = self.language_embeddings[lang_id]
        for entry in seq:
            word = entry.word
            char_emb, _ = self.character_network.compute_embeddings(word, runtime=runtime, language_embeddings=lang_emb)
            if sys.version_info[0] == 2:
                holistic_word = word.decode('utf-8').lower()
            else:
                holistic_word = word.lower()
            if holistic_word in self.encodings.word2int:
                hol_emb = self.holistic_word_embedding[self.encodings.word2int[holistic_word]]
            else:
                hol_emb = self.holistic_word_embedding[self.encodings.word2int['<UNK>']]
            proj_hol = self.hol_proj_w.expr(update=True) * hol_emb
            proj_char = self.char_proj_w.expr(update=True) * char_emb
            proj_lang = self.lang_proj_w.expr(update=True) * lang_emb
            # x_list.append(dy.tanh(proj_char + proj_emb + proj_hol))

            if runtime:
                x_list.append(dy.tanh(proj_char + proj_hol + proj_lang))
            else:
                p1 = random.random()
                p2 = random.random()
                p3 = random.random()
                p4 = random.random()
                m1 = 1
                m2 = 1
                m3 = 1

                if p1 < self.config.input_dropout_prob:
                    m1 = 0
                if p2 < self.config.input_dropout_prob:
                    m2 = 0
                if p3 < self.config.input_dropout_prob:
                    m3 = 0

                scale = 1.0
                if m1 + m2 + m3 > 0:
                    scale = float(3) / (m1 + m2 + m3)
                m1 = dy.scalarInput(m1)
                m2 = dy.scalarInput(m2)
                m3 = dy.scalarInput(m3)
                scale = dy.scalarInput(scale)
                x_list.append(dy.tanh((proj_char * m1 + proj_hol * m2 + proj_lang * m3) * scale))

        # BDLSTM
        rnn_outputs = []
        for fw, bw, dropout in zip(self.bdrnn_fw, self.bdrnn_bw, self.config.layer_dropouts):
            if not runtime:
                fw.set_dropouts(0, dropout)
                bw.set_dropouts(0, dropout)
            else:
                fw.set_dropouts(0, 0)
                bw.set_dropouts(0, 0)
            fw_list = fw.initial_state().transduce(x_list)
            bw_list = list(reversed(bw.initial_state().transduce(reversed(x_list))))
            x_list = [dy.concatenate([x_fw, x_bw, lang_emb]) for x_fw, x_bw in zip(fw_list, bw_list)]
            # if runtime:
            #    x_out = x_list
            # else:
            #    x_out = [dy.dropout(x, dropout) for x in x_list]
            rnn_outputs.append(x_list)

        # SOFTMAX
        mlp_output = []
        for x in rnn_outputs[-1]:
            pre_softmax = []
            for iMLP in range(3):
                mlp_w = self.mlps[iMLP][0]
                mlp_b = self.mlps[iMLP][1]
                inp = x
                for w, b, drop, in zip(mlp_w, mlp_b, self.config.presoftmax_mlp_dropouts):
                    inp = dy.tanh(w.expr(update=True) * inp + b.expr(update=True))
                    if not runtime:
                        inp = dy.dropout(inp, drop)
                pre_softmax.append(inp)
            mlp_output.append(pre_softmax)

        for softmax_inp, aux_softmax_inp in zip(mlp_output, rnn_outputs[self.config.aux_softmax_layer - 1]):
            softmax_list.append([dy.softmax(
                self.softmax_upos_w.expr(update=True) * softmax_inp[0] + self.softmax_upos_b.expr(update=True)),
                dy.softmax(
                    self.softmax_xpos_w.expr(update=True) * softmax_inp[1] + self.softmax_xpos_b.expr(
                        update=True)),
                dy.softmax(
                    self.softmax_attrs_w.expr(update=True) * softmax_inp[
                        2] + self.softmax_attrs_b.expr(update=True))])
            aux_softmax_list.append(
                [dy.softmax(self.aux_softmax_upos_w.expr(update=True) * aux_softmax_inp + self.aux_softmax_upos_b.expr(
                    update=True)),
                 dy.softmax(self.aux_softmax_xpos_w.expr(update=True) * aux_softmax_inp + self.aux_softmax_xpos_b.expr(
                     update=True)),
                 dy.softmax(
                     self.aux_softmax_attrs_w.expr(update=True) * aux_softmax_inp + self.aux_softmax_attrs_b.expr(
                         update=True))])

        return softmax_list, aux_softmax_list

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)

    def tag_sequences(self, sequences):
        new_sequences = []
        for sequence, lang_id in sequences:
            new_sequence = copy.deepcopy(sequence)
            predicted_tags = self.tag(new_sequence, lang_id=lang_id)
            for entryIndex, pred in enumerate(predicted_tags):
                new_sequence[entryIndex].upos = pred[0]
                new_sequence[entryIndex].xpos = pred[1]
                new_sequence[entryIndex].attrs = pred[2]
            new_sequences.append(new_sequence)
        return new_sequences



#
# Author: Tiberiu Boros
#
# Copyright (c) 2019 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Code adapted from https://github.com/rguthrie3/BiLSTM-CRF/blob/master/model.py
# and from https://github.com/neulab/cmu-ner/blob/master/models/decoders.py

import dynet as dy
import numpy as np


class CRFDecoder:
    def __init__(self, model, src_output_dim, tag_emb_dim, tag_size, constraints=None):
        self.model = model
        self.start_id = tag_size
        self.end_id = tag_size + 1
        self.tag_size = tag_size + 2
        tag_size = tag_size + 2

        # optional: transform the hidden space of src encodings into the tag embedding space
        self.W_src2tag_readout = model.add_parameters((tag_emb_dim, src_output_dim))
        self.b_src2tag_readout = model.add_parameters((tag_emb_dim))
        self.b_src2tag_readout.zero()

        self.W_scores_readout2tag = model.add_parameters((tag_size, tag_emb_dim))
        self.b_scores_readout2tag = model.add_parameters((tag_size))
        self.b_scores_readout2tag.zero()

        # (to, from), trans[i] is the transition score to i
        init_transition_matrix = np.random.randn(tag_size, tag_size)  # from, to
        # init_transition_matrix[self.start_id, :] = -1000.0
        # init_transition_matrix[:, self.end_id] = -1000.0
        init_transition_matrix[self.end_id, :] = -1000.0
        init_transition_matrix[:, self.start_id] = -1000.0
        if constraints is not None:
            init_transition_matrix = self._constrained_transition_init(init_transition_matrix, constraints)
        # print init_transition_matrix
        self.transition_matrix = model.lookup_parameters_from_numpy(init_transition_matrix)

        self.interpolation = True  # args.interp_crf_score
        if self.interpolation:
            self.W_weight_transition = model.add_parameters((1, tag_emb_dim))
        self.b_weight_transition = model.add_parameters((1))
        self.b_weight_transition.zero()

    def learn(self, src_enc, tgt_tags):
        return self.decode_loss(src_enc, [tgt_tags])

    def tag(self, src_enc):
        return self.decoding(src_enc)[1]

    def _constrained_transition_init(self, transition_matrix, contraints):
        '''
        :param transition_matrix: numpy array, (from, to)
        :param contraints: [[from_indexes], [to_indexes]]
        :return: newly initialized transition matrix
        '''
        for cons in contraints:
            transition_matrix[cons[0], cons[1]] = -1000.0
        return transition_matrix

    def _log_sum_exp_dim_0(self, x):
        # numerically stable log_sum_exp
        dims = x.dim()
        max_score = dy.max_dim(x, 0)  # (dim_1, batch_size)
        if len(dims[0]) == 1:
            max_score_extend = max_score
        else:
            max_score_reshape = dy.reshape(max_score, (1, dims[0][1]), batch_size=dims[1])
            max_score_extend = dy.concatenate([max_score_reshape] * dims[0][0])
        x = x - max_score_extend
        exp_x = dy.exp(x)
        # (dim_1, batch_size), if no dim_1, return ((1,), batch_size)
        log_sum_exp_x = dy.log(dy.mean_dim(exp_x, d=[0], b=False) * dims[0][0])
        return log_sum_exp_x + max_score

    def forward_alg(self, tag_scores):
        ''' Forward DP for CRF.
        tag_scores (list of batched dy.Tensor): (tag_size, batchsize)
        '''
        # Be aware: if a is lookup_parameter with 2 dimension, then a[i] returns one row;
        # if b = dy.parameter(a), then b[i] returns one column; which means dy.parameter(a) already transpose a
        transpose_transition_score = self.transition_matrix#.expr(update=True)
        # transpose_transition_score = dy.transpose(transition_score)
        # alpha(t', s) = the score of sequence from t=0 to t=t' in log space
        # np_init_alphas = -100.0 * np.ones((self.tag_size, batch_size))
        # np_init_alphas[self.start_id, :] = 0.0
        # alpha_tm1 = dy.inputTensor(np_init_alphas, batched=True)

        alpha_tm1 = transpose_transition_score[self.start_id] + tag_scores[0]
        # self.transition_matrix[i]: from i, column
        # transpose_score[i]: to i, row
        # transpose_score: to, from

        for tag_score in tag_scores[1:]:
            # extend for each transit <to>
            alpha_tm1 = dy.concatenate_cols([alpha_tm1] * self.tag_size)  # (from, to, batch_size)
            # each column i of tag_score will be the repeated emission score to tag i
            tag_score = dy.transpose(dy.concatenate_cols([tag_score] * self.tag_size))
            alpha_t = alpha_tm1 + transpose_transition_score + tag_score
            alpha_tm1 = self._log_sum_exp_dim_0(alpha_t)  # (tag_size, batch_size)

        terminal_alpha = self._log_sum_exp_dim_0(alpha_tm1 + self.transition_matrix[self.end_id])  # (1, batch_size)
        return terminal_alpha

    def score_one_sequence(self, tag_scores, tags, batch_size):
        ''' tags: list of tag ids at each time step '''
        # print tags, batch_size
        # print batch_size
        # print "scoring one sentence"
        tags = [[self.start_id] * batch_size] + tags  # len(tag_scores) = len(tags) - 1
        score = dy.inputTensor(np.zeros(batch_size), batched=True)
        # tag_scores = dy.concatenate_cols(tag_scores) # tot_tags, sent_len, batch_size
        # print "tag dim: ", tag_scores.dim()
        for i in range(len(tags) - 1):
            score += dy.pick_batch(dy.lookup_batch(self.transition_matrix, tags[i + 1]), tags[i]) \
                     + dy.pick_batch(tag_scores[i], tags[i + 1])
        score += dy.pick_batch(dy.lookup_batch(self.transition_matrix, [self.end_id] * batch_size), tags[-1])
        return score

    def _transpose_input(self, seq, padding_token=0):
        # input seq: list of samples [[w1, w2, ..], [w1, w2, ..]]
        max_len = max([len(sent) for sent in seq])
        seq_pad = []
        seq_mask = []
        for i in range(max_len):
            pad_temp = [sent[i] if i < len(sent) else padding_token for sent in seq]
            mask_temp = [1.0 if i < len(sent) else 0.0 for sent in seq]
            seq_pad.append(pad_temp)
            seq_mask.append(mask_temp)

        return seq_pad, seq_mask

    def decode_loss(self, src_encodings, tgt_tags):
        # This is the batched version which requires bucketed batch input with the same length.
        '''
        The length of src_encodings and tgt_tags are time_steps.
        src_encodings: list of dynet.Tensor (src_output_dim, batch_size)
        tgt_tags: list of tag ids [(1, batch_size)]
        return: average of negative log likelihood
        '''
        # TODO: transpose tgt tags first
        batch_size = len(tgt_tags)
        tgt_tags, tgt_mask = self._transpose_input(tgt_tags, 0)
        W_src2tag_readout = self.W_src2tag_readout.expr(update=True)
        b_src2tag_readout = self.b_src2tag_readout.expr(update=True)
        W_score_tag = self.W_scores_readout2tag.expr(update=True)
        b_score_tag = self.b_scores_readout2tag.expr(update=True)

        tag_embs = [dy.tanh(dy.affine_transform([b_src2tag_readout, W_src2tag_readout, src_encoding])) for src_encoding
                    in src_encodings]
        if self.interpolation:
            W_transit = self.W_weight_transition.expr(update=True)
            b_transit = self.b_weight_transition.expr(update=True)
            step_weight_on_transit = [dy.logistic(dy.affine_transform([b_transit, W_transit, tag_emb])) for tag_emb in
                                      tag_embs]

        tag_scores = [dy.affine_transform([b_score_tag, W_score_tag, tag_emb]) for tag_emb in tag_embs]

        # scores over all paths, all scores are in log-space
        forward_scores = self.forward_alg(tag_scores)
        gold_score = self.score_one_sequence(tag_scores, tgt_tags, batch_size)
        # negative log likelihood
        loss = dy.sum_batches(forward_scores - gold_score) / batch_size
        return loss  # , dy.sum_batches(forward_scores)/batch_size, dy.sum_batches(gold_score) / batch_size

    def get_crf_scores(self, src_encodings):
        W_src2tag_readout = self.W_src2tag_readout.expr(update=True)
        b_src2tag_readout = self.b_src2tag_readout.expr(update=True)
        W_score_tag = self.W_scores_readout2tag.expr(update=True)
        b_score_tag = self.b_scores_readout2tag.expr(update=True)

        tag_embs = [dy.tanh(dy.affine_transform([b_src2tag_readout, W_src2tag_readout, src_encoding]))
                    for src_encoding in src_encodings]
        tag_scores = [dy.affine_transform([b_score_tag, W_score_tag, tag_emb]) for tag_emb in tag_embs]

        transpose_transition_score = self.transition_matrix#.expr(update=True)  # (to, from)

        return transpose_transition_score.npvalue(), [ts.npvalue() for ts in tag_scores]

    def decoding(self, src_encodings):
        ''' Viterbi decoding for a single sequence. '''
        W_src2tag_readout = self.W_src2tag_readout.expr(update=True)
        b_src2tag_readout = self.b_src2tag_readout.expr(update=True)
        W_score_tag = self.W_scores_readout2tag.expr(update=True)
        b_score_tag = self.b_scores_readout2tag.expr(update=True)

        tag_embs = [dy.tanh(dy.affine_transform([b_src2tag_readout, W_src2tag_readout, src_encoding]))
                    for src_encoding in src_encodings]
        tag_scores = [dy.affine_transform([b_score_tag, W_score_tag, tag_emb]) for tag_emb in tag_embs]

        back_trace_tags = []
        np_init_alpha = np.ones(self.tag_size) * -2000.0
        np_init_alpha[self.start_id] = 0.0
        max_tm1 = dy.inputTensor(np_init_alpha)
        transpose_transition_score = self.transition_matrix#.expr(update=True) # (to, from)

        for i, tag_score in enumerate(tag_scores):
            max_tm1 = dy.concatenate_cols([max_tm1] * self.tag_size)
            max_t = max_tm1 + transpose_transition_score
            if i != 0:
                eval_score = max_t.npvalue()[:-2, :]
            else:
                eval_score = max_t.npvalue()
            best_tag = np.argmax(eval_score, axis=0)
            back_trace_tags.append(best_tag)
            max_tm1 = dy.inputTensor(eval_score[best_tag, range(self.tag_size)]) + tag_score

        terminal_max_T = max_tm1 + self.transition_matrix[self.end_id]
        eval_terminal = terminal_max_T.npvalue()[:-2]
        best_tag = np.argmax(eval_terminal, axis=0)
        best_path_score = eval_terminal[best_tag]

        best_path = [best_tag]
        for btpoint in reversed(back_trace_tags):
            best_tag = btpoint[best_tag]
            best_path.append(best_tag)
        start = best_path.pop()
        assert start == self.start_id
        best_path.reverse()
        return best_path_score, best_path

    def cal_accuracy(self, pred_path, true_path):
        return np.sum(np.equal(pred_path, true_path).astype(np.float32)) / len(pred_path)


class CRFLabeler:
    def __init__(self, tagset_size, num_lstm_layers, hidden_dim, input_dim, model=None):
        if model is None:
            self.model = dy.Model()
        else:
            self.model = model
        self.tagset_size = tagset_size + 2

        self.START = tagset_size
        self.STOP = tagset_size + 1

        # LSTM parameters
        self.bi_lstm = dy.BiRNNBuilder(num_lstm_layers, input_dim, hidden_dim, self.model, dy.LSTMBuilder)

        # Matrix that maps from Bi-LSTM output to num tags
        self.lstm_to_tags_params = self.model.add_parameters((self.tagset_size, hidden_dim))
        self.lstm_to_tags_bias = self.model.add_parameters(self.tagset_size)
        self.mlp_out = self.model.add_parameters((self.tagset_size, self.tagset_size))
        self.mlp_out_bias = self.model.add_parameters(self.tagset_size)

        # Transition matrix for tagging layer, [i,j] is score of transitioning to i from j
        self.transitions = self.model.add_lookup_parameters((self.tagset_size, self.tagset_size))

    def learn(self, input, tags):
        return self._neg_log_loss(input, tags)

    def tag(self, input):
        return self._viterbi_decoding(self._build_tagging_graph(input))[0]

    def set_dropout(self, p):
        self.bi_lstm.set_dropout(p)

    def disable_dropout(self):
        self.bi_lstm.disable_dropout()

    def _build_tagging_graph(self, sentence):
        # embeddings = [self.word_rep(w) for w in sentence]
        embeddings = sentence

        lstm_out = self.bi_lstm.transduce(embeddings)

        H = self.lstm_to_tags_params.expr(update=True)
        Hb = self.lstm_to_tags_bias.expr(update=True)
        O = self.mlp_out.expr(update=True)
        Ob = self.mlp_out_bias.expr(update=True)
        # H = dy.parameter(self.lstm_to_tags_params)
        # Hb = dy.parameter(self.lstm_to_tags_bias)
        # O = dy.parameter(self.mlp_out)
        # Ob = dy.parameter(self.mlp_out_bias)
        scores = []
        for rep in lstm_out:
            score_t = O * dy.tanh(H * rep + Hb) + Ob
            scores.append(score_t)

        return scores

    def _score_sentence(self, observations, tags):
        assert len(observations) == len(tags)
        score_seq = [0]
        score = dy.scalarInput(0)
        tags = [self.START] + tags
        for i, obs in enumerate(observations):
            score = score + dy.pick(self.transitions[tags[i + 1]], tags[i]) + dy.pick(obs, tags[i + 1])
            score_seq.append(score.value())
        score = score + dy.pick(self.transitions[self.STOP], tags[-1])
        return score

    def _viterbi_loss(self, observations, tags):
        # observations = self.build_tagging_graph(sentence)
        viterbi_tags, viterbi_score = self._viterbi_decoding(observations)
        if viterbi_tags != tags:
            gold_score = self._score_sentence(observations, tags)
            return (viterbi_score - gold_score), viterbi_tags
        else:
            return dy.scalarInput(0), viterbi_tags

    def _neg_log_loss(self, sentence, tags):
        observations = self._build_tagging_graph(sentence)
        gold_score = self._score_sentence(observations, tags)
        forward_score = self._forward(observations)
        return forward_score - gold_score

    def _forward(self, observations):

        def log_sum_exp(scores):
            npval = scores.npvalue()
            argmax_score = np.argmax(npval)
            max_score_expr = dy.pick(scores, argmax_score)
            max_score_expr_broadcast = dy.concatenate([max_score_expr] * self.tagset_size)
            return max_score_expr + dy.log(dy.sum_dim(dy.transpose(dy.exp(scores - max_score_expr_broadcast)), [1]))

        init_alphas = [-1e10] * self.tagset_size
        init_alphas[self.START] = 0
        for_expr = dy.inputVector(init_alphas)
        for obs in observations:
            alphas_t = []
            for next_tag in range(self.tagset_size):
                obs_broadcast = dy.concatenate([dy.pick(obs, next_tag)] * self.tagset_size)
                next_tag_expr = for_expr + self.transitions[next_tag] + obs_broadcast
                alphas_t.append(log_sum_exp(next_tag_expr))
            for_expr = dy.concatenate(alphas_t)
        terminal_expr = for_expr + self.transitions[self.STOP]
        alpha = log_sum_exp(terminal_expr)
        return alpha

    def _viterbi_decoding(self, observations):
        backpointers = []
        init_vvars = [-1e10] * self.tagset_size
        init_vvars[self.START] = 0  # <Start> has all the probability
        for_expr = dy.inputVector(init_vvars)
        trans_exprs = [self.transitions[idx] for idx in range(self.tagset_size)]
        for obs in observations:
            bptrs_t = []
            vvars_t = []
            for next_tag in range(self.tagset_size):
                next_tag_expr = for_expr + trans_exprs[next_tag]
                next_tag_arr = next_tag_expr.npvalue()
                best_tag_id = np.argmax(next_tag_arr)
                bptrs_t.append(best_tag_id)
                vvars_t.append(dy.pick(next_tag_expr, best_tag_id))
            for_expr = dy.concatenate(vvars_t) + obs
            backpointers.append(bptrs_t)
        # Perform final transition to terminal
        terminal_expr = for_expr + trans_exprs[self.STOP]
        terminal_arr = terminal_expr.npvalue()
        best_tag_id = np.argmax(terminal_arr)
        path_score = dy.pick(terminal_expr, best_tag_id)
        # Reverse over the backpointers to get the best path
        best_path = [best_tag_id]  # Start with the tag that was best for terminal
        for bptrs_t in reversed(backpointers):
            best_tag_id = bptrs_t[best_tag_id]
            best_path.append(best_tag_id)
        start = best_path.pop()  # Remove the start symbol
        best_path.reverse()
        assert start == self.START
        # Return best path and best path's score
        return best_path, path_score



# -*- coding: utf-8 -*-

import sys
import os

# Append parent dir to sys path.
os.sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from cube.io_utils.model_store import ModelMetadata, ModelStore
from datetime import datetime

if __name__ == "__main__":        
    if len(sys.argv)!=2:
        print("Usage: python3 import_model.py path-to-my-model")
        print("Example: 'python3 import_model.py my_model-1.0.zip")
        sys.exit(0)
        
    model_path = sys.argv[1]
    
    if not os.path.exists(model_path):
        raise Exception("File {} not found!".format(model_path))
       
    if ".zip" not in model_path:
        raise Exception("File {} must be a zip! See export_model.py and the online tutorial on how to package a locally trained model!".format(model_path))
    
    model_store_object = ModelStore()
    
    model_store_object.import_local_model(model_path)
   


#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import os
import yaml
import string
import requests
import tarfile
from tqdm import tqdm

sys.path.append('')
from typing import Optional, Union
from cube.io_utils.objects import Document, Word, Token, Sentence
from cube.io_utils.encodings import Encodings
from cube.io_utils.config import CompoundConfig, TokenizerConfig, ParserConfig, LemmatizerConfig
from cube.networks.compound import Compound
from cube.networks.parser import Parser
from cube.networks.tokenizer import Tokenizer
from cube.networks.lemmatizer import Lemmatizer
from pathlib import Path
from cube.networks.lm import LMHelperHF, LMHelperFT
from cube.networks.utils_tokenizer import TokenCollateHF, TokenCollateFTLanguasito
from cube.networks.utils import MorphoCollate, Word2TargetCollate


class CubeObj:
    def __init__(self, model_base: str, device: str = 'cpu', lang: str = None):
        self._cwe = None
        # word expander
        path = '{0}-trf-cwe'.format(model_base)
        if os.path.exists('{0}.best'.format(path)):
            config = CompoundConfig(filename='{0}.config'.format(path))
            encodings = Encodings()
            encodings.load('{0}.encodings'.format(path))
            self._cwe = Compound(config, encodings)
            self._cwe.load('{0}.best'.format(path))
            self._cwe.to(device)

        # tokenizer
        path = '{0}-trf-tokenizer'.format(model_base)
        g_conf = yaml.safe_load(open('{0}.yaml'.format(path)))
        self._lang2id = {}
        for lng in g_conf['language_codes']:
            self._lang2id[lng] = len(self._lang2id)
        self._default_lang_id = self._lang2id[g_conf['language_map'][lang]]
        self._default_lang = lang
        config = TokenizerConfig(filename='{0}.config'.format(path))
        print(self._default_lang_id)
        lm_model = config.lm_model
        encodings = Encodings()
        encodings.load('{0}.encodings'.format(path))
        if lm_model.startswith('transformer'):
            self._tokenizer_collate = TokenCollateHF(encodings,
                                                     lm_device=device,
                                                     lm_model=lm_model.split(':')[-1],
                                                     no_space_lang=config.no_space_lang,
                                                     lang_id=self._default_lang_id)
        else:
            self._tokenizer_collate = TokenCollateFTLanguasito(encodings,
                                                               lm_device=device,
                                                               lm_model=lm_model,
                                                               no_space_lang=config.no_space_lang,
                                                               lang_id=self._default_lang_id)


        self._tokenizer = Tokenizer(config, encodings, language_codes=g_conf['language_codes'],
                                    ext_word_emb=self._tokenizer_collate.get_embeddings_size())
        self._tokenizer.load('{0}.best'.format(path))
        self._tokenizer.to(device)

        # lemmatizer
        path = '{0}-trf-lemmatizer'.format(model_base)
        config = LemmatizerConfig(filename='{0}.config'.format(path))
        encodings = Encodings()
        encodings.load('{0}.encodings'.format(path))
        self._lemmatizer = Lemmatizer(config, encodings)
        self._lemmatizer.load('{0}.best'.format(path))
        self._lemmatizer.to(device)
        self._lemmatizer_collate = Word2TargetCollate(encodings)
        # parser-tagger
        path = '{0}-trf-parser'.format(model_base)
        config = ParserConfig(filename='{0}.config'.format(path))
        lm_model = config.lm_model
        if lm_model.startswith('transformer'):
            self._lm_helper = LMHelperHF(model=lm_model.split(':')[-1])
        else:
            self._lm_helper = LMHelperFT(model=lm_model.split(':')[-1])

        encodings = Encodings()
        encodings.load('{0}.encodings'.format(path))
        self._parser = Parser(config, encodings, language_codes=g_conf['language_codes'],
                              ext_word_emb=self._lm_helper.get_embedding_size())
        self._parser.load('{0}.best'.format(path))
        self._parser.to(device)
        self._parser_collate = MorphoCollate(encodings)

    def __call__(self, text: Union[str, Document], flavour: Optional[str] = None):
        lang_id = self._default_lang_id
        if flavour is not None:
            if flavour not in self._lang2id:
                print("Unsupported language flavour")
                print("Please choose from: {0}".format(' '.join([k for k in self._lang2id])))
                raise Exception("Unsupported language flavour\nPlease choose from: {0}".
                                format(' '.join([k for k in self._lang2id])))
            lang_id = self._lang2id[flavour]
        if isinstance(text, str):
            doc = self._tokenizer.process(text, self._tokenizer_collate, lang_id=lang_id, num_workers=0)
            if self._cwe is not None:
                doc = self._cwe.process(doc, self._lemmatizer_collate, num_workers=0)
        else:
            doc = text

        self._lm_helper.apply(doc)
        self._parser.process(doc, self._parser_collate, num_workers=0)
        self._lemmatizer.process(doc, self._lemmatizer_collate, num_workers=0)
        for seq in doc.sentences:
            for w in seq.words:
                if w.upos =='PUNCT':
                    w.lemma = w.word
        return doc


def _download_file(url: str, filename: str, description=None):
    r = requests.get(url, stream=True)
    if r.status_code != 200:
        raise Exception(f"Error getting {url}, received status_code {r.status_code}")
    file_size = int(r.headers['Content-Length'])
    chunk_size = 1024

    with open(filename, 'wb') as fp:
        with tqdm(total=file_size, unit='B', unit_scale=True, desc=description, unit_divisor=1024,
                  disable=True if description is None else False, leave=False) as progressbar:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk is not None:
                    fp.write(chunk)
                    fp.flush()
                    progressbar.update(len(chunk))

    return r.status_code


def _download_model(local_path, lang):
    download_base = "https://github.com/adobe/NLP-Cube-Models/raw/3.0/models/{0}.tar.gz-a".format(lang)
    file_base = "{0}.tar.gz-a".format(lang)
    terminations = string.ascii_lowercase[:20]
    file_list = []
    for t in terminations:
        download_url = '{0}{1}'.format(download_base, t)
        target_file = str(os.path.join(local_path, file_base))
        target_file = '{0}{1}'.format(target_file, t)
        try:
            if _download_file(download_url, target_file, description='Part {0}'.format(t)) != 200:
                break
        except:
            break
        file_list.append(target_file)

    target_file = os.path.join(local_path, file_base[:-2])

    f_out = open(target_file, 'wb')
    for file in file_list:
        f_in = open(file, 'rb')
        while True:
            buffer = f_in.read(1024 * 1024)
            if not buffer:
                break
            f_out.write(buffer)
    f_out.close()

    tar = tarfile.open(target_file, 'r:gz')
    tar.extractall(local_path)
    tar.close()


def _load(lang: str, device: Optional[str] = 'cpu') -> CubeObj:
    try:
        local_user_home = str(Path.home())
        local_user_storage = os.path.join(local_user_home, '.nlpcube', '3.0')
        os.makedirs(local_user_storage, exist_ok=True)
        lang_path = os.path.join(local_user_storage, lang)
        if not os.path.exists(lang_path):
            _download_model(local_user_storage, lang)

        return CubeObj('{0}/{1}'.format(lang_path, lang), device=device, lang=lang)
    except:
        raise Exception("There was a problem retrieving this language. Either it is unsupported or your Internet "
                        "connection is down.\n\nTo check for supported languages, visit "
                        "https://github.com/adobe/NLP-Cube/\n\nIt is hard to maintain models for all UD Treebanks. "
                        "This is way we are only including a handful of"
                        "languages with the official distribution. "
                        "However, we can include additional languages upon request"
                        "\n\nTo make a request for supporting a new language please create an issue on GitHub")


class Cube:
    def __init__(self, verbose=False):
        self._instance = None

    def load(self, lang: str, device: Optional[str] = 'cpu'):
        self._instance = _load(lang, device)

    def __call__(self, text: Union[str, Document], flavour: Optional[str] = None):
        return self._instance(text, flavour=flavour)



# -*- coding: utf-8 -*-

import sys
import os

# Append parent dir to sys path.
os.sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from cube.io_utils.model_store import ModelMetadata, ModelStore
from datetime import datetime

if __name__ == "__main__":    
    print("Usage: python3 export_model.py path-to-my-model --tokenizer(optional) --compound-word-expander(optional) --lemmatizer(optional) --tagger(optional) --parser(optional)")
    print("Example: 'python3 export_model.py path-to-my-model --tokenizer --tagger' will create a zip file named 'language_code-model_version.zip' (taken from the metadata.json) containing a tokenizer and a tagger.")
       
    # parameter checking
    _tokenizer = False
    _compound_word_expander = False  
    _lemmatizer = False  
    _parser = False  
    _tagger = False 
    model_folder_path = ""
    
    for param in sys.argv:
        if not param.startswith("--"):
            model_folder_path = param
        else:
            if "--tokenizer" in param:
                _tokenizer = True
            if "--compound-word-expander" in param:
                _compound_word_expander = True
            if "--lemmatizer" in param:
                _lemmatizer = True
            if "--tagger" in param:
                _tagger = True
            if "--parser" in param:
                _parser = True                    
    
    print("\n\tModel folder: "+model_folder_path)
    print("\tUse tokenizer: {}".format(_tokenizer))
    print("\tUse compound word expander: {}".format(_compound_word_expander))
    print("\tUse lemmatizer: {}".format(_lemmatizer))
    print("\tUse tagger: {}".format(_tagger))
    print("\tUse parser: {}\n".format(_parser))
    
    # check if path exists
    if not os.path.exists(model_folder_path):
        raise Exception ("Model folder not found!")
        
    # check if metadata exists
    if not os.path.exists(os.path.join(model_folder_path,"metadata.json")):
        raise Exception ("metadata.json not found in model folder!")
    
    # check if metadata is valid
    metadata = ModelMetadata()
    metadata.read(os.path.join(model_folder_path, "metadata.json"))
    
    output_folder_path = os.path.dirname(model_folder_path)    
    model_store_object = ModelStore(disk_path=output_folder_path)
    
    model_store_object.package_model(model_folder_path, output_folder_path, metadata, should_contain_tokenizer = _tokenizer, should_contain_compound_word_expander = _compound_word_expander, should_contain_lemmatizer = _lemmatizer, should_contain_tagger = _tagger, should_contain_parser = _parser)


#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import re
from cube.misc.misc import fopen


class Encodings(object):

    def __init__(self, verbose=True):
        self.word_list = {}
        self.hol_word_list = []
        self.char2int = {}
        self.label2int = {}
        self.labels = []
        self.word2int = {}
        self.upos2int = {}
        self.xpos2int = {}
        self.attrs2int = {}
        self.upos_list = []
        self.xpos_list = []
        self.attrs_list = []
        self.characters = []
        self.verbose = verbose

    def compute(self, train, dev, tag_type=None, word_cutoff=7, char_cutoff=5, CUPT_format=False):
        if self.verbose:
            sys.stdout.write("Computing encoding maps... ")
            sys.stdout.flush()

        self.word2int['<PAD>'] = 0
        self.hol_word_list.append('<PAD>')
        self.word2int['<UNK>'] = 1
        self.hol_word_list.append('<UNK>')
        self.char2int['<PAD>'] = 0
        self.char2int['<UNK>'] = 1
        self.char2int[' '] = 2
        self.upos2int['<PAD>'] = 0
        self.upos_list.append('<PAD>')
        self.xpos2int['<PAD>'] = 0
        self.xpos_list.append('<PAD>')
        self.attrs2int['<PAD>'] = 0
        self.attrs_list.append('<PAD>')
        self.upos2int['<UNK>'] = 1
        self.upos_list.append('<UNK>')
        self.xpos2int['<UNK>'] = 1
        self.xpos_list.append('<UNK>')
        self.attrs2int['<UNK>'] = 1
        self.attrs_list.append('<PAD>')
        self.label2int['<PAD>'] = 0
        self.labels.append('<PAD>')
        self.label2int['<UNK>'] = 1
        self.labels.append('<UNK>')

        self.characters.append("<PAD>")
        self.characters.append("<UNK>")
        self.characters.append(" ")
        char_count = {}
        word_count = {}
        for [seq, lang_id] in train.sequences:
            for entry in seq:
                word = entry.word.lower()
                if word not in word_count:
                    word_count[word] = 1
                else:
                    word_count[word] = word_count[word] + 1
                if word not in self.word_list:
                    self.word_list[word] = 0  # word is inside trainset

                uniword = entry.word.lower()
                uniword = re.sub('\d', '0', uniword)
                for i in range(len(uniword)):
                    char = uniword[i].lower()
                    if char not in char_count:
                        char_count[char] = 1
                    else:
                        char_count[char] = char_count[char] + 1

                        # if char not in self.char2int:
                        #    self.char2int[char] = len(self.char2int)

                label = entry.label

                if CUPT_format and tag_type == 'label':
                    if entry.label != "*":
                        labels = entry.label.split(';')
                        entry_labels = [label.split(':')[1] for label in labels if ':' in label]
                        for entry_label in entry_labels:
                            self.label2int.setdefault(entry_label, len(self.label2int))
                else:
                    if label not in self.label2int:
                        self.label2int[label] = len(self.label2int)
                        self.labels.append(label)

                # morphological encodings
                if entry.upos not in self.upos2int:
                    self.upos2int[entry.upos] = len(self.upos2int)
                    self.upos_list.append(entry.upos)
                if entry.xpos not in self.xpos2int:
                    self.xpos2int[entry.xpos] = len(self.xpos2int)
                    self.xpos_list.append(entry.xpos)
                if entry.attrs not in self.attrs2int:
                    self.attrs2int[entry.attrs] = len(self.attrs2int)
                    self.attrs_list.append(entry.attrs)

        for [seq, lang_id] in dev.sequences:
            for entry in seq:
                word = entry.word.lower()
                if word not in self.word_list:
                    self.word_list[word] = 1  # word is inside devset only

        for word in word_count:
            if word_count[word] >= word_cutoff:
                self.word2int[word] = len(self.word2int)
                self.hol_word_list.append(word)
        for char in char_count:
            if char_count[char] >= char_cutoff and char not in self.char2int:
                self.char2int[char] = len(self.char2int)
                self.characters.append(char)

        # force add digits
        for digit in range(10):
            ds = str(digit)
            if ds not in self.char2int:
                self.char2int[ds] = len(self.char2int)
                self.characters.append(ds)
        if self.verbose:
            sys.stdout.write("done\n")

            print("Unique words: " + str(len(self.word_list)))
            print("Unique chars: " + str(len(self.char2int)))
            print("Unique labels: " + str(len(self.label2int)))
            print("Unique UPOS: " + str(len(self.upos2int)))
            print("Unique XPOS: " + str(len(self.xpos2int)))
            print("Unique ATTRS: " + str(len(self.attrs2int)))
            print("Holistic word count: " + str(len(self.word2int)))

    def update_wordlist(self, dataset):
        for seq in dataset.sequences:
            for entry in seq:
                word = entry.word.lower()
                if word not in self.word_list:
                    self.word_list[word] = 2  # word is inside an auxiliarly set (probably test)

    def load(self, filename):
        # We only read character2int, labels, holistic words and label2int here. word_list should be recomputed for every dataset (if deemed necessary)
        with fopen(filename, "r") as f:
            line = f.readline()

            num_labels = int(line.split(" ")[1])
            if self.verbose:
                print("Loading labels " + str(num_labels))
            self.labels = [""] * num_labels
            for _ in range(num_labels):
                line = f.readline()
                parts = line.split("\t")
                key = parts[0]
                value = int(parts[1])
                self.label2int[key] = value
                self.labels[value] = key

            line = f.readline()
            num_characters = int(line.split(" ")[1])
            self.characters = [""] * num_characters
            if self.verbose:
                print("Loading characters " + str(num_characters))
            for _ in range(num_characters):
                line = f.readline()
                parts = line.split("\t")
                key = parts[0]
                value = int(parts[1])
                self.char2int[key] = value
                self.characters[value] = key
            line = f.readline()
            num_words = int(line.split(" ")[1])
            if self.verbose:
                print("Loading words " + str(num_words))
            for _x in range(num_words):
                line = f.readline()
                parts = line.split("\t")
                key = parts[0]
                value = int(parts[1])
                self.word2int[key] = value

            # morphological attributes
            line = f.readline()
            num_labels = int(line.split(" ")[1])
            if self.verbose:
                print("Loading upos " + str(num_labels))
            self.upos_list = [""] * num_labels
            for _ in range(num_labels):
                line = f.readline()
                parts = line.split("\t")
                key = parts[0]
                value = int(parts[1])
                self.upos2int[key] = value
                self.upos_list[value] = key

            line = f.readline()
            num_labels = int(line.split(" ")[1])
            self.xpos_list = [""] * num_labels
            if self.verbose:
                print("Loading xpos " + str(num_labels))
            for _ in range(num_labels):
                line = f.readline()
                parts = line.split("\t")
                key = parts[0]
                value = int(parts[1])
                self.xpos2int[key] = value
                self.xpos_list[value] = key

            line = f.readline()
            num_labels = int(line.split(" ")[1])
            self.attrs_list = [""] * num_labels
            if self.verbose:
                print("Loading attrs " + str(num_labels))
            for _ in range(num_labels):
                line = f.readline()
                parts = line.split("\t")
                key = parts[0]
                value = int(parts[1])
                self.attrs2int[key] = value
                self.attrs_list[value] = key
            f.close()

    def save(self, filename):
        f = fopen(filename, "w")
        f.write("LABELS " + str(len(self.label2int)) + "\n")
        for label in self.label2int:
            f.write(str(label) + "\t" + str(self.label2int[label]) + "\n")
        f.write("CHARACTERS " + str(len(self.char2int)) + "\n")
        for character in self.char2int:
            f.write(character + "\t" + str(self.char2int[character]) + "\n")
        f.write("WORDS " + str(len(self.word2int)) + "\n")
        for word in self.word2int:
            f.write(word + "\t" + str(self.word2int[word]) + "\n")

        f.write("UPOS " + str(len(self.upos2int)) + "\n")
        for label in self.upos2int:
            f.write(label + "\t" + str(self.upos2int[label]) + "\n")
        f.write("XPOS " + str(len(self.xpos2int)) + "\n")
        for label in self.xpos2int:
            f.write(label + "\t" + str(self.xpos2int[label]) + "\n")
        f.write("ATTRS " + str(len(self.attrs2int)) + "\n")
        for label in self.attrs2int:
            f.write(label + "\t" + str(self.attrs2int[label]) + "\n")
        f.close()



import os, sys, yaml

sys.path.append(".")
from audioop import add

from pytorch_lightning.callbacks import EarlyStopping

from argparse import ArgumentParser

from torch.utils.data import DataLoader
import pytorch_lightning as pl

from cube.io_utils.config import TaggerConfig, ParserConfig, TokenizerConfig, LemmatizerConfig, CompoundConfig
from cube.io_utils.encodings import Encodings
from cube.io_utils.objects import Document
from cube.networks.tokenizer import Tokenizer
from cube.networks.tagger import Tagger
from cube.networks.parser import Parser
from cube.networks.lemmatizer import Lemmatizer
from cube.networks.compound import Compound
from cube.networks.utils import MorphoDataset, MorphoCollate, TokenizationDataset, \
    Word2TargetCollate, LemmaDataset, CompoundDataset
from cube.networks.utils_tokenizer import TokenCollateHF, TokenCollateFTLanguasito
from cube.networks.lm import LMHelperFT, LMHelperHF, LMHelperLanguasito


class Trainer():
    def __init__(self, task: str, language_map: {}, language_codes: [], train_files: {}, dev_files: {}, test_files: {},
                 args):
        self.task = None
        if task not in ["tokenizer", "lemmatizer", "cwe", "tagger", "parser"]:
            raise Exception("Task must be one of: tokenizer, lemmatizer, cwe, tagger or parser.")

        self.store_prefix = args.store
        self.language_map = language_map
        self.language_codes = language_codes
        self.args = args

        # TODO assert lang_id matches train
        # TODO assert train files code are found in dev files
        # lang_code_list = sorted(list(set(code for code in train_files )))

        self.task = task
        self.doc_train = Document()
        self.doc_dev = Document()
        self.doc_test = Document()

        for lang_code in train_files:
            print("Reading train file for language code {} : {}".format(lang_code, train_files[lang_code]))
            self.doc_train.load(train_files[lang_code], lang_id=language_codes.index(lang_code))
        for lang_code in dev_files:
            print("Reading dev file for language code {} : {}".format(lang_code, dev_files[lang_code]))
            self.doc_dev.load(dev_files[lang_code], lang_id=language_codes.index(lang_code))
        for lang_code in test_files:
            print("Reading test file for language code {} : {}".format(lang_code, test_files[lang_code]))
            self.doc_test.load(test_files[lang_code], lang_id=language_codes.index(lang_code))

        # ensure store dir exists
        i = self.store_prefix.rfind("/")
        if i > 0:
            if i >= len(self.store_prefix) - 1:
                raise Exception(
                    "store_prefix is a folder; please specify the prefix of the models after the '/', like 'data/tagger'.")

            target_folder = self.store_prefix[:i]
            model_prefix = self.store_prefix[i + 1:]
            os.makedirs(target_folder, exist_ok=True)
            print("Saving model in {}, with prefix {}".format(target_folder, model_prefix))
        else:
            print("Saving model in the current folder, with prefix {}".format(self.store_prefix))

    def fit(self):
        if self.task not in ["tokenizer", "lemmatizer", "cwe", "tagger", "parser"]:
            raise Exception("Task must be one of: tokenizer, lemmatizer, cwe, tagger or parser.")

        with open(self.args.store + ".yaml", 'w') as f:
            yaml.dump({"language_map": self.language_map, "language_codes": self.language_codes}, f, sort_keys=True)

        enc = Encodings()
        enc.compute(self.doc_train, None)
        enc.save('{0}.encodings'.format(self.store_prefix))

        if self.task == "tokenizer":
            config = TokenizerConfig()
            no_space_lang = Tokenizer._detect_no_space_lang(self.doc_train)
            print("NO_SPACE_LANG = " + str(no_space_lang))
            config.no_space_lang = no_space_lang
        if self.task == "tagger":
            config = TaggerConfig()
        if self.task == "lemmatizer":
            config = LemmatizerConfig()
        if self.task == "parser":
            config = ParserConfig()
        if self.task == "cwe":
            config = CompoundConfig()
        config.lm_model = self.args.lm_model
        if self.args.config_file:
            config.load(self.args.config_file)
            if self.args.lm_model is not None:
                config.lm_model = self.args.lm_model
        config.save('{}.config'.format(self.args.store))

        if self.task != "tokenizer" and self.task != 'lemmatizer' and self.task != 'cwe':
            lm_model = config.lm_model
            parts = lm_model.split(':')
            if parts[0] not in ['transformer', 'fasttext', 'languasito']:
                print("Error: model prefix should be in the form of transformer: fasttext: or languasito:")
                sys.exit(0)
            if parts[0] == 'transformer':
                helper = LMHelperHF(device=self.args.lm_device, model=parts[1])
            elif parts[0] == 'fasttext':
                helper = LMHelperFT(device=self.args.lm_device, model=parts[1])
            elif parts[0] == 'languasito':
                helper = LMHelperLanguasito(device=self.args.lm_device, model=parts[1])
            helper.apply(self.doc_dev)
            helper.apply(self.doc_train)

        if self.task == "tokenizer":
            trainset = TokenizationDataset(self.doc_train)
            devset = TokenizationDataset(self.doc_dev, shuffle=False)
        elif self.task == 'parser' or self.task == 'tagger':
            trainset = MorphoDataset(self.doc_train)
            devset = MorphoDataset(self.doc_dev)
        elif self.task == 'lemmatizer':
            trainset = LemmaDataset(self.doc_train)
            devset = LemmaDataset(self.doc_dev)
        elif self.task == 'cwe':
            trainset = CompoundDataset(self.doc_train)
            devset = CompoundDataset(self.doc_dev)

        collate = MorphoCollate(enc)

        # per task specific settings
        callbacks = []
        if self.task == "tokenizer":
            early_stopping_callback = EarlyStopping(
                monitor='val/early_meta',
                patience=args.patience,
                verbose=True,
                mode='max'
            )
            parts = args.lm_model.split(':')
            if parts[0] == 'transformer':
                collate = TokenCollateHF(enc, lm_device=args.lm_device, lm_model=parts[1],
                                         no_space_lang=config.no_space_lang)
            else:
                collate = TokenCollateFTLanguasito(enc, lm_device=args.lm_device, lm_model=args.lm_model,
                                                   no_space_lang=config.no_space_lang)

            callbacks = [early_stopping_callback, Tokenizer.PrintAndSaveCallback(self.store_prefix)]
            model = Tokenizer(config=config, encodings=enc, language_codes=self.language_codes,
                              ext_word_emb=collate.get_embeddings_size(), max_seq_len=collate.max_seq_len)

        if self.task == "tagger":
            early_stopping_callback = EarlyStopping(
                monitor='val/early_meta',
                patience=args.patience,
                verbose=True,
                mode='max'
            )
            callbacks = [early_stopping_callback, Tagger.PrintAndSaveCallback(self.store_prefix)]
            model = Tagger(config=config, encodings=enc, language_codes=self.language_codes,
                           ext_word_emb=helper.get_embedding_size())

        if self.task == "parser":
            collate = MorphoCollate(enc, add_parsing=True, rhl_win_size=config.rhl_win_size)
            early_stopping_callback = EarlyStopping(
                monitor='val/early_meta',
                patience=args.patience,
                verbose=True,
                mode='max'
            )
            callbacks = [early_stopping_callback, Parser.PrintAndSaveCallback(self.store_prefix)]
            model = Parser(config=config, encodings=enc, language_codes=self.language_codes,
                           ext_word_emb=helper.get_embedding_size())

        if self.task == "lemmatizer":
            collate = Word2TargetCollate(enc)
            early_stopping_callback = EarlyStopping(
                monitor='val/early_meta',
                patience=args.patience,
                verbose=True,
                mode='max'
            )
            callbacks = [early_stopping_callback, Lemmatizer.PrintAndSaveCallback(self.store_prefix)]
            model = Lemmatizer(config=config, encodings=enc, language_codes=self.language_codes)

        if self.task == "cwe":
            collate = Word2TargetCollate(enc)
            early_stopping_callback = EarlyStopping(
                monitor='val/early_meta',
                patience=args.patience,
                verbose=True,
                mode='max'
            )
            callbacks = [early_stopping_callback, Compound.PrintAndSaveCallback(self.store_prefix)]
            model = Compound(config=config, encodings=enc, language_codes=self.language_codes)
            # extra check to see if there is actually any compound in this language
            if len(trainset._examples) == 0 or len(devset._examples) == 0:
                print("\nTrain/dev data for this language does not contain any compound words; there is nothing to train.")
                return

        # dataloaders
        train_loader = DataLoader(trainset, batch_size=self.args.batch_size, collate_fn=collate.collate_fn,
                                  shuffle=True,
                                  num_workers=self.args.num_workers)
        val_loader = DataLoader(devset, batch_size=self.args.batch_size, collate_fn=collate.collate_fn,
                                num_workers=self.args.num_workers)

        # pre-train checks
        resume_from_checkpoint = None
        if self.args.resume is True:
            resume_from_checkpoint = self.store_prefix + ".last"
            if not os.path.exists(resume_from_checkpoint):
                raise Exception("Resume from checkpoint: {} not found!".format(resume_from_checkpoint))

        """if self.args.gpus == 0:
            acc = 'ddp_cpu'
        else:
            acc = 'ddp'
        """

        trainer = pl.Trainer(
            gpus=args.gpus,
            accelerator=args.accelerator,
            #num_nodes=1,
            default_root_dir='data/',
            callbacks=callbacks,
            resume_from_checkpoint=resume_from_checkpoint,
            accumulate_grad_batches=args.accumulate_grad_batches,
            # limit_train_batches=100,
            # limit_val_batches=4,
        )

        # run fit
        print("\nStarting train\n")
        trainer.fit(model, train_loader, val_loader)


if __name__ == '__main__':
    parser = ArgumentParser(description='NLP-Cube Trainer Helper')
    parser.add_argument('--task', action='store', dest='task',
                        help='Type of task : "tokenizer", "lemmatizer", "cwe", "tagger", "parser"')
    parser.add_argument('--train', action='store', dest='train_file',
                        help='Start building a tagger model')
    parser.add_argument('--patience', action='store', type=int, default=20, dest='patience',
                        help='Number of epochs before early stopping (default=20)')
    parser.add_argument('--store', action='store', dest='store', help='Output base', default='data/model')
    parser.add_argument('--num-workers', action='store', dest='num_workers', type=int,
                        help='How many dataloaders to use (default=4)', default=4)
    parser.add_argument('--batch-size', action='store', type=int, default=16, dest='batch_size',
                        help='Batch size (default=16)')
    parser.add_argument('--debug', action='store_true', dest='debug',
                        help='Do some standard stuff to debug the model')
    parser.add_argument('--resume', action='store_true', dest='resume', help='Resume training')
    parser.add_argument('--lm-model', action='store', dest='lm_model',
                        help='What LM model to use (default=xlm-roberta-base)')
    parser.add_argument('--lm-device', action='store', dest='lm_device', default='cuda:0',
                        help='Where to load LM (default=cuda:0)')
    parser.add_argument('--config', action='store', dest='config_file', help='Load config file')

    parser = pl.Trainer.add_argparse_args(parser) # add all pytorch lightning params here as well

    args = parser.parse_args()

    with open(args.train_file) as file:
        train_config = yaml.full_load(file)

    trainer_object = Trainer(
        task=args.task,
        language_map=train_config["language_map"],
        language_codes=train_config["language_codes"],
        train_files=train_config["train_files"],
        dev_files=train_config["dev_files"],
        test_files=train_config["test_files"],
        args=args,
    )

    trainer_object.fit()




import dynet as dy
import numpy as np


def orthonormal_VanillaLSTMBuilder(lstm_layers, input_dims, lstm_hiddens, pc):
    builder = dy.VanillaLSTMBuilder(lstm_layers, input_dims, lstm_hiddens, pc)
    for layer, params in enumerate(builder.get_parameters()):
        W = orthonormal_initializer(lstm_hiddens, lstm_hiddens + (lstm_hiddens if layer > 0 else input_dims))
        W_h, W_x = W[:, :lstm_hiddens], W[:, lstm_hiddens:]
        params[0].set_value(np.concatenate([W_x] * 4, 0))
        params[1].set_value(np.concatenate([W_h] * 4, 0))
        b = np.zeros(4 * lstm_hiddens, dtype=np.float32)
        b[lstm_hiddens:2 * lstm_hiddens] = -1.0
        params[2].set_value(b)
    return builder


def orthonormal_initializer(output_size, input_size):
    """
    adopted from Deng Cai
    adopted from Timothy Dozat https://github.com/tdozat/Parser/blob/master/lib/linalg.py
    """
    print (output_size, input_size)
    I = np.eye(output_size)
    lr = .1
    eps = .05 / (output_size + input_size)
    success = False
    tries = 0
    while not success and tries < 10:
        Q = np.random.randn(input_size, output_size) / np.sqrt(output_size)
        for i in range(100):
            QTQmI = Q.T.dot(Q) - I
            loss = np.sum(QTQmI ** 2 / 2)
            Q2 = Q ** 2
            Q -= lr * Q.dot(QTQmI) / (
                        np.abs(Q2 + Q2.sum(axis=0, keepdims=True) + Q2.sum(axis=1, keepdims=True) - 1) + eps)
            if np.max(Q) > 1e6 or loss > 1e6 or not np.isfinite(loss):
                tries += 1
                lr /= 2
                break
        success = True
    if success:
        #print('Orthogonal pretrainer loss: %.2e' % loss)
        pass
    else:
        #print('Orthogonal pretrainer failed, using non-orthogonal random matrix')
        Q = np.random.randn(input_size, output_size) / np.sqrt(output_size)
    return np.transpose(Q.astype(np.float32))



import json, os, yaml, uuid
import ntpath
from shutil import rmtree, copyfile
import logging
from tqdm.autonotebook import tqdm as tqdm

parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
os.sys.path.insert(0, parent_dir)

from cube.io_utils.modelstore import ModelStore

if __name__ == "__main__":
    # set vars here
    VERSION = "1.0"
    FOLDER_WITH_YAMLS = os.path.abspath("scripts//train//2.7//language//")
    FOLDER_WITH_TRAINED_MODELS = "/media/echo/5CA436CBA436A802/work/models"  # os.path.abspath("models")
    FOLDER_WHERE_TO_OUTPUT_EVERYTHING = "/media/echo/5CA436CBA436A802/work/nlp-cube-models"  # os.path.abspath("nlp-cube-models")
    URL_ROOT_FOR_MODELS = "https://raw.githubusercontent.com/adobe/NLP-Cube-Models/3.0/models/" # !! make sure it ends with a /

    """    
    0. Open existing catalog, create new key "version" if it does not exist
    
    1. Load all yaml files:
        lang_code_id : "lang_code" -> index in original yaml file
        lang_map : "name" -> "lang_code"        
    
    2. For each lang_code, check to see if files are available to form a model. If not, report errors.
    
        3. Copy all files in temp folder
    
        4. Pack all files and split in shards
    
        5. Move in final model dir with name lang_code.version.###
    
        6. For all entries in lang_map, if lang_code is the current one, create entry in catalog if does not exist
            "name":["version","link","lang_code_index"] 
    
    
    7. Write catalog 
    """

    logger = logging.getLogger("cube")
    log_handler = logging.StreamHandler()
    log_formatter = logging.Formatter(
        fmt="[%(levelname)8s | %(asctime)s | %(filename)-20s:%(lineno)3s | %(funcName)-26s] %(message)s",
        datefmt='%Y-%m-%d %H:%M:%S')
    log_handler.setFormatter(log_formatter)
    logger.addHandler(log_handler)
    logger.setLevel(logging.DEBUG)



    # STEP 0 - open current catalog or create new
    catalog_path = os.path.join(FOLDER_WHERE_TO_OUTPUT_EVERYTHING, "catalog.json")
    if os.path.exists(catalog_path):
        catalog = json.load(open(catalog_path, "r", encoding="utf8"))
    else:
        catalog = {}


    # STEP 1 - load all yamls
    langcode_to_index = {}
    name_to_langcode = {}

    yamls = os.listdir(FOLDER_WITH_YAMLS) # this just lists the files without path
    yamls = [os.path.abspath(os.path.join(FOLDER_WITH_YAMLS, x)) for x in yamls] # fill full path in
    yamls = [x for x in yamls if os.path.isfile(x) and x.endswith(".yaml")] # filter out possible junk
    print("I see {} yamls in {}.".format(len(yamls), FOLDER_WITH_YAMLS))

    for yaml_file in yamls:
        y = yaml.safe_load(open(yaml_file, "r"))

        for index, lang_code in enumerate(y["language_codes"]):
            if lang_code in langcode_to_index:
                print(f"Warning! lang_code {lang_code} already found in langcode_to_index!")
            langcode_to_index[lang_code] = index

        for name in y["language_map"]:
            if name in name_to_langcode:
                print(f"Warning! name {name} already found in name_to_langcode!")
            name_to_langcode[name] = y["language_map"][name]

    # STEP 2 - for each lang_code
    for lang_code in langcode_to_index:
        print(f"Working on {lang_code} ...")

        # STEP 3 - copy all relevant files for this language_code in temp folder
        files = os.listdir(FOLDER_WITH_TRAINED_MODELS)  # this just lists the files without path
        files = [os.path.abspath(os.path.join(FOLDER_WITH_TRAINED_MODELS, x)) for x in files]  # fill full path in
        files = [x for x in files if os.path.isfile(x) and ".last" not in x]

        # get major language_code
        tt = [x for x in files if lang_code in x]
        if len(tt) == 0:
            print(f"\tCould not get major language code!")
            continue
        tt = tt[0]
        tt = ntpath.basename(tt)
        major_lang_code = tt.split("-")[0]
        print(f"\t major language code is [{major_lang_code}]")

        valid_files = []
        # copy encodings and config
        for f in files:
            ff = ntpath.basename(f)
            if ff.startswith(major_lang_code+"-") and (".config" in ff or ".encodings" in ff):
                valid_files.append(f)
        # copy lang_codes with tok, best and las
        for f in files:
            if lang_code in f and (".tok" in f or ".best" in f or ".las" in f):
                valid_files.append(f)
        files = valid_files

        # check they are valid
        found_tokenizer, found_lemmatizer, found_parser = False, False, False
        for f in files:
            if "tokenizer" in f:
                found_tokenizer = True
            if "lemmatizer" in f:
                found_lemmatizer = True
            if "parser" in f:
                found_parser = True

        if not(found_tokenizer and found_lemmatizer and found_parser):
            print(f"\t {lang_code} does not have all files: tokenizer={found_tokenizer}, lemmatizer={found_lemmatizer}, parser={found_parser}, skipping")
            with open("log.txt", "a") as f:
                f.write(f"\t {lang_code} does not have all files: tokenizer={found_tokenizer}, lemmatizer={found_lemmatizer}, parser={found_parser}, skipping\n")
            continue

        temp_folder = os.path.join(FOLDER_WHERE_TO_OUTPUT_EVERYTHING, str(uuid.uuid4().hex))
        os.mkdir(temp_folder)

        # copy files to temp folder
        print("\t copying files to temp folder ... ")
        for src_file in files:
            _, name = os.path.split(src_file)
            dst_file = os.path.join(temp_folder, name)
            # print((src_file, dst_file))
            copyfile(src_file, dst_file)

        # pack folder in zip file
        zip_file_path = os.path.join(temp_folder, lang_code + "-" + VERSION)
        split_count = ModelStore._pack_model(
            input_folder=temp_folder,
            output_folder=FOLDER_WHERE_TO_OUTPUT_EVERYTHING,
            model_name=lang_code + "-" + VERSION,
            split_size_in_mb=99)

        # delete temp folder
        print("\t deleting temp folder ...")
        if os.path.exists(temp_folder):
            rmtree(temp_folder, ignore_errors=True)

        # STEP 7 - make a catalog entry for all language names affected :
        entry = {
            "version": VERSION,
            "link": URL_ROOT_FOR_MODELS + lang_code + "-" + VERSION,
            "langid": langcode_to_index[lang_code],
            "parts": split_count
        }

        for name in name_to_langcode:
            if name_to_langcode[name] == lang_code:
                print(f"\t making a catalog entry for [{name}] -> [{lang_code}], {split_count} parts, langid {langcode_to_index[lang_code]}")
                if name not in catalog:
                    catalog[name] = []

                catalog[name].append(entry)

    print("Finished processing all language codes, writing catalog ... ")

    json.dump(catalog, open(catalog_path, "w", encoding="utf8"), indent=4, sort_keys=True)

    print("Done.")



import os, sys, logging, json, requests, uuid, shutil
import zipfile
from pathlib import Path
from typing import Optional, List, Tuple
from tqdm.autonotebook import tqdm as tqdm

logger = logging.getLogger('cube')

class ModelStore:
    """
    The purpose of this class is to be called from the api.load(language) and give back a list of component paths,
    or download them if they are not locally stored

    """
    root_path = os.path.join(str(Path.home()), ".nlpcube")
    catalog_url = "https://raw.githubusercontent.com/adobe/NLP-Cube-Models/3.0/models/catalog.json"

    @staticmethod
    def solve(lang: str, version: str = "latest", check_for_latest: bool = False) -> Tuple[dict, int]:
        """
        TODO docs
        :param lang: Language
        :param version: force particular version, else latest
        :return: Dict of paths, language id
        """
        paths = {
            "tokenizer": {},
            "cwe": {},
            "lemmatizer": {},
            "tagger": {},
            "parser": {}
        }

        # check local catalog and download if not present
        catalog = ModelStore._get_catalog(check_for_latest = check_for_latest)

        # identify all entries that match this lang and select the appropriate one
        if lang not in catalog:
            raise Exception(f"Language '{lang}' is not available!")

        entries = catalog[lang]
        if len(entries) == 0:
            raise Exception(f"Language '{lang}' is not available!")

        found = False
        if version != "latest": # check for specific version
            for entry in entries:
                if entry["version"] == version:
                    found = True
                    model_url = entry["link"]
                    langid = entry["langid"]
                    parts = entry["parts"]
        else:
            max_version = 0.0
            for entry in entries: # check for latest version
                try: # get version
                    entry_version = float(entry["version"])
                    if entry_version > max_version:
                        found = True
                        max_version = entry_version
                        model_url = entry["link"]
                        langid = entry["langid"]
                        parts = entry["parts"]
                except:
                    pass
            version = max_version

        if not found:
            raise Exception(f"Language '{lang}', version '{version}' not found, but there are other versions available!")

        # check if it's present, else download model in folder "name.version", return abspath of model folder
        model_folder = ModelStore._get_model(model_url, parts)

        # compose and return paths
        files = [x for x in os.listdir(model_folder)] # file names only

        #files = [os.path.abspath(os.path.join(model_folder, x)) for x in files]  # get full path

        tokenizer_files = [x for x in files if "tokenizer" in x]
        cwe_files = [x for x in files if "cwe" in x]
        lemmatizer_files = [x for x in files if "lemmatizer" in x]
        tagger_files = [x for x in files if "tagger" in x]
        parser_files = [x for x in files if "parser" in x]

        if len(tokenizer_files) > 0:
            tokenizer_entry = {"config": ModelStore.__get_file_path(tokenizer_files, ".config", model_folder),
                               "encodings": ModelStore.__get_file_path(tokenizer_files, ".encodings", model_folder),
                               #"sent": ModelStore.__get_file_path(tokenizer_files, ".sent", model_folder),
                               "model": ModelStore.__get_file_path(tokenizer_files, ".tok", model_folder)}

            if tokenizer_entry["config"] and tokenizer_entry["encodings"] and tokenizer_entry["model"]:
                paths["tokenizer"] = tokenizer_entry

        if len(cwe_files) > 0:
            cwe_entry = {"config": ModelStore.__get_file_path(cwe_files, ".config", model_folder),
                         "encodings": ModelStore.__get_file_path(cwe_files, ".encodings", model_folder),
                         "model": ModelStore.__get_file_path(cwe_files, ".best", model_folder)}

            if cwe_entry["config"] and cwe_entry["encodings"] and cwe_entry["model"]:
                paths["cwe"] = cwe_entry

        if len(lemmatizer_files) > 0:
            lemmatizer_entry = {"config": ModelStore.__get_file_path(lemmatizer_files, ".config", model_folder),
                                "encodings": ModelStore.__get_file_path(lemmatizer_files, ".encodings", model_folder),
                                "model": ModelStore.__get_file_path(lemmatizer_files, ".best", model_folder)}

            if lemmatizer_entry["config"] and lemmatizer_entry["encodings"] and lemmatizer_entry["model"]:
                paths["lemmatizer"] = lemmatizer_entry

        """
        if len(tagger_files) > 0:
            tagger_entry = {"config": ModelStore.__get_file_path(tagger_files, ".config", model_folder),
                            "encodings": ModelStore.__get_file_path(tagger_files, ".encodings", model_folder),
                            "sent": ModelStore.__get_file_path(tagger_files, ".sent", model_folder),
                            "tok": ModelStore.__get_file_path(tagger_files, ".tok", model_folder)}

            if tagger_entry["config"] and tagger_entry["encodings"] and tagger_entry["sent"] and \
                    tagger_entry["tok"]:
                paths["tagger"] = tagger_entry
        """

        if len(parser_files) > 0:
            parser_entry = {"config": ModelStore.__get_file_path(parser_files, ".config", model_folder),
                            "encodings": ModelStore.__get_file_path(parser_files, ".encodings", model_folder),
                            "model": ModelStore.__get_file_path(parser_files, ".las", model_folder)}

            if parser_entry["config"] and parser_entry["encodings"] and parser_entry["model"]:
                paths["parser"] = parser_entry

        return paths, langid

    @staticmethod
    def __get_file_path(files, extension, model_folder):
        """
        This function returns the abspath of the only one file from the 'files' list that has the given extension.
        'files' does not contain paths, only filenames
        """
        valid_files = []
        for file in files:
            if file.endswith(extension):
                valid_files.append(file)
        if len(valid_files)!=1:
            return None
        return os.path.abspath(os.path.join(model_folder, valid_files[0]))


    @staticmethod
    def _get_catalog(check_for_latest: bool = False):
        local_path = os.path.join(ModelStore.root_path, "catalog.json")

        if not os.path.exists(local_path) or check_for_latest is True:
            print("Catalog either does not exist or looking for updates, downloading ... ")
            status_code = ModelStore.__download_file(ModelStore.catalog_url, local_path)
            if status_code != 200:
                raise Exception(
                    "Catalog download failed with status_code {}".format(status_code))

        if not os.path.exists(local_path):
            raise Exception("Sanity check failed, catalog file not found locally, though it was downloaded!")

        with open(local_path, "r", encoding="utf8") as f:
            catalog = json.load(f)

        return catalog

    @staticmethod
    def _get_model(model_url: str, parts: int):
        if model_url.endswith("/"): # remove last /,
            model_url = model_url[:-1]

        model_name = model_url.split("/")[-1] # e.g. en_partut-1.0
        model_folder = os.path.join(ModelStore.root_path, "models", model_name)

        # check if model already exists
        if os.path.exists(model_folder):
            if len([f for f in os.listdir(model_folder) if os.path.isfile(os.path.join(model_folder, f))])>0:
                return os.path.abspath(model_folder) # model is present
                # todo Sanity check that model is valid here, not in solve.

        temp_folder = os.path.join(ModelStore.root_path, str(uuid.uuid4().hex))
        os.mkdir(temp_folder)

        # download each file
        print("Downloading model {} ...".format(model_name))
        current_part = 0
        for current_part in range(parts):
            current_file = os.path.join(temp_folder, "{}.{}".format(model_name, current_part))
            current_url = "{}.{}".format(model_url, current_part)
            status_code = ModelStore.__download_file(current_url, current_file, description=f"    ... download model part {current_part+1}/{parts}")
            if status_code != 200:
                raise Exception(f"Error downloading file {current_url}, received status code {status_code}")

        print("Merging model parts ...")
        zip_file = os.path.join(temp_folder, "archive.zip")
        with open(zip_file, "wb") as f:
            for i in range(parts):
                with open(os.path.join(temp_folder, "{}.{}".format(model_name, i)), "rb") as r:
                    f.write(r.read())

        print("Unzipping ...")
        os.makedirs(model_folder, exist_ok=True)
        zip = zipfile.ZipFile(zip_file, "r")
        zip.extractall(model_folder)
        zip.close()

        print("Cleaning up ...")
        if os.path.exists(temp_folder) and os.path.isdir(temp_folder):
            shutil.rmtree(temp_folder)

        # todo sanity check
        print("Model downloaded successfully!")

        return os.path.abspath(model_folder)

    @staticmethod
    def __download_file(url: str, filename: str, description=None):
        r = requests.get(url, stream=True)
        if r.status_code != 200:
            raise Exception(f"Error getting {url}, received status_code {r.status_code}")
        file_size = int(r.headers['Content-Length'])
        chunk_size = 1024

        with open(filename, 'wb') as fp:
            with tqdm(total=file_size, unit='B', unit_scale=True, desc=description, unit_divisor=1024, disable= True if description is None else False, leave = False) as progressbar:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk is not None:
                        fp.write(chunk)
                        fp.flush()
                        progressbar.update(len(chunk))

        return r.status_code

    @staticmethod
    def _pack_model(input_folder: str, output_folder: str, model_name:str, split_size_in_mb: int = 50) -> int:
        """
        Zips everything in input folder, splits it in output_folder with model_name.<part##>, return number of parts
        """

        # zip all files in input_folder as input_folder/archive.zip
        print(f"Zipping files from {input_folder}:")
        zip_file_path = os.path.join(input_folder, "archive.zip")
        zip = zipfile.ZipFile(zip_file_path, 'w', compression=zipfile.ZIP_DEFLATED)
        root_len = len(os.path.abspath(input_folder))
        for root, dirs, files in os.walk(input_folder):
            archive_root = os.path.abspath(root)[root_len:]
            for f in files:
                if ".zip" in f:
                    continue
                fullpath = os.path.join(root, f)
                archive_name = os.path.join(archive_root, f)
                print(f"\t adding {fullpath} ...")
                zip.write(fullpath, archive_name, zipfile.ZIP_DEFLATED)

        zip.close()

        # split archive.zip in shards in output_folder
        counter = 0
        with open(zip_file_path, "rb") as f:
            while True:
                byte_s = f.read(split_size_in_mb * 1024 * 1024)
                if not byte_s:
                    break
                chunk_file_path = os.path.join(output_folder, "{}.{}".format(model_name, counter))
                logging.info("\t writing {:.2f}MB to {} ...".format(len(byte_s) / 1024 / 1024, chunk_file_path))
                with open(chunk_file_path, "wb") as r:
                    r.write(byte_s)
                counter += 1

        # return number of files
        return counter




# ensure we get a valid root path for local model storage
ModelStore.root_path = os.path.join(str(Path.home()), ".nlpcube")
if not os.path.exists(ModelStore.root_path):
    os.makedirs(ModelStore.root_path, exist_ok=True)



from cube.data.objects import Doc, Sentence, Word, Token


class ComponentType:
    Tokenizer = 1
    CWExpander = 2
    POSTagger = 3
    Lemmatizer = 4
    Parser = 5
    NER = 6


class Model:
    def __init__(self):
        pass

    def __call__(self, input_object, **kwargs):
        return None


class Component():
    def __init__(self, use_gpu: bool = True, gpu_batch_size: int = 1):
        self.input_format = Doc  # type of data object or str
        self.output_format = Doc  # type of data object
        self.depends = []  # list of other components
        self.provides = []  # list of other components

        self.model_filepath = None

        self.use_gpu = use_gpu
        self.gpu_batch_size = gpu_batch_size

        self.model = None

    def load_model(self, model_path):
        pass

    def process(self, input_object):
        assert (self.model is not None), "Model is none, please load model first"
        return self.model(input_object=input_object)


class TokenizerComponent(Component):
    def __init__(self):
        super().__init__()
        self.input_format = str
        self.depends = []
        self.provides = [ComponentType.Tokenizer]


class CWExpanderComponent(Component):
    def __init__(self):
        super().__init__()
        self.depends = [ComponentType.Tokenizer]
        self.provides = [ComponentType.CWExpander]


class POSTaggerComponent(Component):
    def __init__(self):
        super().__init__()
        self.depends = [ComponentType.Tokenizer, ComponentType.CWExpander]
        self.provides = [ComponentType.POSTagger]


class LemmatizerComponent(Component):
    def __init__(self):
        super().__init__()
        self.depends = [ComponentType.Tokenizer, ComponentType.CWExpander, ComponentType.Parser]
        self.provides = [ComponentType.Lemmatizer]


class ParserComponent(Component):
    def __init__(self):
        super().__init__()
        self.depends = [ComponentType.Tokenizer, ComponentType.CWExpander]
        self.provides = [ComponentType.Parser]


class NERComponent(Component):
    def __init__(self):
        super().__init__()
        self.depends = [ComponentType.Tokenizer, ComponentType.CWExpander, ComponentType.Parser]
        self.provides = [ComponentType.NER]


class Pipeline():
    def is_valid(components: Component):
        available = set()
        required = set()
        for component in components:
            available |= set(component.provides)
            required |= set(component.depends)
        return required.issubset(available)



# -*- coding: utf-8 -*-

import sys
import os

class Cube(object):
    def __init__(self, verbose=False, random_seed=None, memory=512, autobatch=False, use_gpu=False):
        """
        Create an empty instance for Cube
        Before it can be used, you must call @method load with @param language_code set to your target language        
        """
        self._loaded = False
        self._verbose = verbose           
        import dynet_config
        
        if random_seed != None:
            if not isinstance(random_seed, int):
                raise Exception ("Random seed must be an integer!")   
            if random_seed == 0:
                print("[Warning] While Python and Numpy's seeds are now set to 0, DyNet uses 0 to reset the seed generator (fully random). Use any non-zero int value to set DyNet to a fixed random seed.")
            # set python random seed
            import random
            random.seed(random_seed)
            #set numpy random seed
            import numpy as np
            np.random.seed(random_seed)
        else:
            random_seed = 0 # this is the default value for DyNet (meaning full random)        
            
        dynet_config.set(mem=memory, random_seed=random_seed, autobatch=autobatch)
        if use_gpu:
            dynet_config.set_gpu()        

    def load(self, language_code, version="latest", local_models_repository=None, local_embeddings_file=None, tokenization=True, compound_word_expanding=False, tagging=True, lemmatization=True, parsing=True):
        """
        Loads the pipeline with all available models for the target language.

        @param lang_code: Target language code. See http://opensource.adobe.com/NLP-Cube/ for available languages and their codes
        @param version: "latest" to get the latest version, or other specific version in like "1.0", "2.1", etc .
       
        """
        from .io_utils.encodings import Encodings
        from .io_utils.embeddings import WordEmbeddings
        from .io_utils.model_store import ModelMetadata, ModelStore
        from .io_utils.config import TieredTokenizerConfig, CompoundWordConfig, LemmatizerConfig, TaggerConfig, ParserConfig
        from .generic_networks.tokenizers import TieredTokenizer
        from .generic_networks.token_expanders import CompoundWordExpander
        from .generic_networks.lemmatizers import FSTLemmatizer
        from .generic_networks.taggers import BDRNNTagger
        from .generic_networks.parsers import BDRNNParser        
        
        self._tokenizer = None  # tokenizer object, default is None
        self._compound_word_expander = None  # compound word expander, default is None
        self._lemmatizer = None  # lemmatizer object, default is None
        self._parser = None  # parser object, default is None
        self._tagger = None  # tagger object, default is None
        self.metadata = ModelMetadata()
        
        # Initialize a ModelStore object
        if local_models_repository: 
            model_store_object = ModelStore(disk_path=local_models_repository)
        else: 
            model_store_object = ModelStore()

        # Find a local model or download it if it does not exist, returning the local model folder path
        model_folder_path = model_store_object.find(lang_code=language_code, version=version, verbose=self._verbose)

        # If the model contains metadata, load it        
        if os.path.isfile(os.path.join(model_folder_path, "metadata.json")): 
            self.metadata.read(os.path.join(model_folder_path, "metadata.json"))
        else:
            self.metadata = None

        # Load embeddings                
        embeddings = WordEmbeddings(verbose=False)
        if self._verbose:
            sys.stdout.write('\tLoading embeddings ... \n')
        if local_embeddings_file is not None:        
            embeddings.read_from_file(local_embeddings_file, None, full_load=False)
        else: # embeddings file is not manually specified
            if self.metadata is None: # no metadata exists
                raise Exception("When using a locally-trained model please specify a path to a local embeddings file (local_embeddings_file cannot be None).")    
            else: # load from the metadata path
                if self.metadata.embeddings_file_name is None or self.metadata.embeddings_file_name == "":
                    # load a dummy embedding
                    embeddings.load_dummy_embeddings()
                else:
                    # load full embedding from file
                    emb_path = os.path.join(model_store_object.embeddings_repository, self.metadata.embeddings_file_name)
                    if not os.path.exists(emb_path):
                        raise Exception("Embeddings file not found: {}".format(emb_path))
                    embeddings.read_from_file(emb_path, None, full_load=False)
       
        # 1. Load tokenizer
        if tokenization:
            if not os.path.isfile(os.path.join(model_folder_path, 'tokenizer-tok.bestAcc')):
                sys.stdout.write('\tTokenization is not available on this model. \n')
            else:
                if self._verbose:
                    sys.stdout.write('\tLoading tokenization model ...\n')
                tokenizer_encodings = Encodings(verbose=False)
                tokenizer_encodings.load(os.path.join(model_folder_path, 'tokenizer.encodings'))
                config = TieredTokenizerConfig(os.path.join(model_folder_path, 'tokenizer.conf'))
                self._tokenizer = TieredTokenizer(config, tokenizer_encodings, embeddings, runtime=True)
                self._tokenizer.load(os.path.join(model_folder_path, 'tokenizer'))

        # 2. Load compound
        if compound_word_expanding:
            if not os.path.isfile(os.path.join(model_folder_path, 'compound.bestAcc')):
                if self._verbose:  # supress warning here because many languages do not have compund words
                    sys.stdout.write('\tCompound word expansion is not available on this model. \n')
            else:
                if self._verbose:
                    sys.stdout.write('\tLoading compound word expander model ...\n')
                compound_encodings = Encodings(verbose=False)
                compound_encodings.load(os.path.join(model_folder_path, 'compound.encodings'))
                config = CompoundWordConfig(os.path.join(model_folder_path, 'compound.conf'))
                self._compound_word_expander = CompoundWordExpander(config, compound_encodings, embeddings,
                                                                    runtime=True)
                self._compound_word_expander.load(os.path.join(model_folder_path, 'compound.bestAcc'))

        # 3. Load lemmatizer
        if lemmatization:
            if not os.path.isfile(os.path.join(model_folder_path, 'lemmatizer.bestAcc')):
                sys.stdout.write('\tLemmatizer is not available on this model. \n')
            else:
                if self._verbose:
                    sys.stdout.write('\tLoading lemmatization model ...\n')
                lemmatizer_encodings = Encodings(verbose=False)
                lemmatizer_encodings.load(os.path.join(model_folder_path, 'lemmatizer.encodings'))
                config = LemmatizerConfig(os.path.join(model_folder_path, 'lemmatizer.conf'))
                self._lemmatizer = FSTLemmatizer(config, lemmatizer_encodings, embeddings, runtime=True)
                self._lemmatizer.load(os.path.join(model_folder_path, 'lemmatizer.bestAcc'))

        # 4. Load tagger
        if tagging or lemmatization:  # we need tagging for lemmatization
            if not os.path.isfile(os.path.join(model_folder_path, 'tagger.bestUPOS')):
                sys.stdout.write('\tTagging is not available on this model. \n')
                if lemmatization:
                    sys.stdout.write('\t\tDisabling the lemmatization model due to missing tagger. \n')
                    self._lemmatizer = None
            else:
                if self._verbose:
                    sys.stdout.write('\tLoading tagger model ...\n')
                tagger_encodings = Encodings(verbose=False)
                tagger_encodings.load(os.path.join(model_folder_path, 'tagger.encodings'))
                config = TaggerConfig(os.path.join(model_folder_path, 'tagger.conf'))
                self._tagger = [None, None, None]
                self._tagger[0] = BDRNNTagger(config, tagger_encodings, embeddings, runtime=True)
                self._tagger[0].load(os.path.join(model_folder_path, 'tagger.bestUPOS'))
                self._tagger[1] = BDRNNTagger(config, tagger_encodings, embeddings, runtime=True)
                self._tagger[1].load(os.path.join(model_folder_path, 'tagger.bestXPOS'))
                self._tagger[2] = BDRNNTagger(config, tagger_encodings, embeddings, runtime=True)
                self._tagger[2].load(os.path.join(model_folder_path, 'tagger.bestATTRS'))

        # 5. Load parser
        if parsing:
            if not os.path.isfile(os.path.join(model_folder_path, 'parser.bestUAS')):
                sys.stdout.write('\tParsing is not available on this model... \n')
            else:
                if self._verbose:
                    sys.stdout.write('\tLoading parser model ...\n')
                parser_encodings = Encodings(verbose=False)
                parser_encodings.load(os.path.join(model_folder_path, 'parser.encodings'))
                config = ParserConfig(os.path.join(model_folder_path, 'parser.conf'))
                self._parser = BDRNNParser(config, parser_encodings, embeddings, runtime=True)
                self._parser.load(os.path.join(model_folder_path, 'parser.bestUAS'))

        self._loaded = True
        if self._verbose:
            sys.stdout.write('Model loading complete.\n\n')

    def __call__(self, text):
        if not self._loaded:
            raise Exception("Cube object is initialized but no model is loaded (eg.: call cube.load('en') )")

        sequences = []
        if self._tokenizer:
            if not isinstance(text, str):
                raise Exception("The text argument must be a string!")
            # split text by lines
            input_lines = text.split("\n")
            for input_line in input_lines:                
                sequences+=self._tokenizer.tokenize(input_line)                
        else:
            if not isinstance(text, list):
                raise Exception("The text argument must be a list of lists of tokens!")
            sequences = text  # the input should already be tokenized

        if self._compound_word_expander:
            sequences = self._compound_word_expander.expand_sequences(sequences)

        if self._parser:
            sequences = self._parser.parse_sequences(sequences)

        if self._tagger or self._lemmatizer:
            import copy
            new_sequences = []
            for sequence in sequences:                
                new_sequence = copy.deepcopy(sequence)                
                predicted_tags_UPOS = self._tagger[0].tag(new_sequence)
                predicted_tags_XPOS = self._tagger[1].tag(new_sequence)
                predicted_tags_ATTRS = self._tagger[2].tag(new_sequence)
                for entryIndex in range(len(new_sequence)):
                    new_sequence[entryIndex].upos = predicted_tags_UPOS[entryIndex][0]
                    new_sequence[entryIndex].xpos = predicted_tags_XPOS[entryIndex][1]
                    new_sequence[entryIndex].attrs = predicted_tags_ATTRS[entryIndex][2]
                new_sequences.append(new_sequence)
            sequences = new_sequences

        if self._lemmatizer:
            sequences = self._lemmatizer.lemmatize_sequences(sequences)

        return sequences


if __name__ == "__main__":
    cube = Cube(verbose=True)
    cube.load('en', tokenization=True, compound_word_expanding=False, tagging=True, lemmatization=True, parsing=True)
    cube.metadata.info()

    text = "I'm a success today because I had a friend who believed in me and I didn't have the heart to let him down. This is a quote by Abraham Lincoln."

    sentences = cube(text)

    for sentence in sentences:
        print()
        for token in sentence:
            line = ""
            line += str(token.index) + "\t"
            line += token.word + "\t"
            line += token.lemma + "\t"
            line += token.upos + "\t"
            line += token.xpos + "\t"
            line += token.attrs + "\t"
            line += str(token.head) + "\t"
            line += token.label + "\t"
            line += token.deps + "\t"
            line += token.space_after
            print(line)



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

sys.path.append("..")
import optparse
from flask import Flask
from flask import Response
from cube.api import Cube

app = Flask(__name__)
singletonServer = False

lang2cube = {}


@app.route('/')
def index():
    return "NLP-Cube server running. Use /help to learn more."


@app.route('/nlp', methods=['GET', 'POST'])
def nlp():
    from flask import request
    if request.args.get('text'):
        query = request.args.get('text')
    else:
        query = request.form.get('text')

    if request.args.get('lang'):
        lang = request.args.get('lang')
    else:
        lang = request.form.get('lang')

    if request.args.get('format'):
        format = request.args.get('format')
    else:
        format = request.form.get('format')

    if not format:
        format = 'CONLL'

    if format != 'CONLL' and format != 'JSON':
        return Response("Allowed values for 'format' are CONLL or JSON", mimetype='text/plain',
                        status=500)
    if not query or not lang:
        return Response("You need to specify the language (lang) and text (text) parameters", mimetype='text/plain',
                        status=500)

    if lang not in lang2cube:
        return Response("This language has not beed preloaded during server startup", mimetype='text/plain',
                        status=500)

    thecube = lang2cube[lang]
    result = thecube(query)
    if format == 'CONLL':
        text = ""
        for seq in result:
            for entry in seq:
                text += str(
                    entry.index) + "\t" + entry.word + "\t" + entry.lemma + "\t" + entry.upos + "\t" + entry.xpos + "\t" + entry.attrs + "\t" + str(
                    entry.head) + "\t" + entry.label + "\t" + entry.deps + "\t" + entry.space_after + "\n"
            text += "\n"
        return Response(text, mimetype='text/plain',
                        status=200)
    else:
        import json
        new_seqs = []
        for seq in result:
            new_seq = []
            for entry in seq:
                new_seq.append(entry.__dict__)
            new_seqs.append(new_seq)
        text = json.dumps(new_seqs, sort_keys=False, indent=4)

        return Response(text, mimetype='application/json',
                        status=200)


@app.route('/help', methods=['GET', 'POST'])
def help():
    text = "NLP-Cube server\n\n" \
           "Use /nlp endpoint to process any text.\n" \
           "\tParameters:\n" \
           "\t\t lang - language code\n" \
           "\t\t text - text to process\n" \
           "\t\t format - output format for data: CONLL|JSON (default is CONLL with plain/text output)"
    return Response(text, mimetype='text/plain',
                    status=200)


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option('--port', action='store', dest='port', type='int', default=8080,
                      help='Binding port for web service (default: 8080)')
    parser.add_option('--host', action='store', dest='host', default='0.0.0.0',
                      help='Binding IP for server (default: 0.0.0.0)')
    parser.add_option('--lang', action='append', dest='languages', default=[],
                      help='Preload language. You can use this param multiple times: --lang en --lang fr ... (default is just ["en"])')

    (params, _) = parser.parse_args(sys.argv)

    if len(params.languages) == 0:
        params.languages = ['en']

    for lang in params.languages:
        lang2cube[lang] = Cube(verbose=True)
        lang2cube[lang].load(lang)

    app.run(port=params.port, host=params.host)



import sys

from pytorch_lightning.overrides.data_parallel import LightningDistributedDataParallel

sys.path.append('')
import os, argparse

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import pytorch_lightning as pl
import torch.nn as nn
import torch.nn.functional as F
import torch
from torch.utils.data.dataset import Dataset
from torch.utils.data import DataLoader
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
import numpy as np
from cube.io_utils.objects import Document, Sentence, Token, Word
from cube.io_utils.encodings import Encodings
from cube.io_utils.config import ParserConfig
import numpy as np
from cube.networks.modules import ConvNorm, LinearNorm, BilinearAttention, Attention, MLP, DeepBiaffine
import random

from cube.networks.utils import MorphoCollate, MorphoDataset, GreedyDecoder, ChuLiuEdmondsDecoder, unpack, mask_concat

from cube.networks.modules import WordGram


class Parser(pl.LightningModule):
    def __init__(self, config: ParserConfig, encodings: Encodings, language_codes: [] = None, ext_word_emb=0):
        super().__init__()
        self._config = config
        self._encodings = encodings
        if not isinstance(ext_word_emb, list):
            ext_word_emb = [ext_word_emb]
        self._ext_word_emb = ext_word_emb

        self._word_net = WordGram(len(encodings.char2int), num_langs=encodings.num_langs + 1,
                                  num_filters=config.char_filter_size, char_emb_size=config.char_emb_size,
                                  lang_emb_size=config.lang_emb_size, num_layers=config.char_layers)
        self._zero_emb = nn.Embedding(1, config.char_filter_size // 2)
        self._num_langs = encodings.num_langs
        self._language_codes = language_codes

        ext2int = []
        for input_size in self._ext_word_emb:
            module = MLP(input_size, config.external_proj_size)
            ext2int.append(module)
        self._ext_proj = nn.ModuleList(ext2int)

        conv_layers = []
        cs_inp = config.char_filter_size // 2 + config.lang_emb_size + config.word_emb_size + config.external_proj_size
        NUM_FILTERS = config.cnn_filter
        for _ in range(config.cnn_layers):
            conv_layer = nn.Sequential(
                ConvNorm(cs_inp,
                         NUM_FILTERS,
                         kernel_size=5, stride=1,
                         padding=2,
                         dilation=1, w_init_gain='tanh'),
                nn.BatchNorm1d(NUM_FILTERS))
            conv_layers.append(conv_layer)
            cs_inp = NUM_FILTERS // 2 + config.lang_emb_size
        self._word_emb = nn.Embedding(len(encodings.word2int), config.word_emb_size, padding_idx=0)
        self._lang_emb = nn.Embedding(encodings.num_langs + 1, config.lang_emb_size, padding_idx=0)
        self._convs = nn.ModuleList(conv_layers)

        self._aupos = LinearNorm(config.char_filter_size // 2 + config.lang_emb_size, len(encodings.upos2int))
        self._axpos = LinearNorm(config.char_filter_size // 2 + config.lang_emb_size, len(encodings.xpos2int))
        self._aattrs = LinearNorm(config.char_filter_size // 2 + config.lang_emb_size, len(encodings.attrs2int))

        self._pre_morpho = LinearNorm(NUM_FILTERS // 2 + config.lang_emb_size, NUM_FILTERS // 2)
        self._upos = LinearNorm(NUM_FILTERS // 2 + config.lang_emb_size, len(encodings.upos2int))
        self._attrs = LinearNorm(64 + NUM_FILTERS // 2 + config.lang_emb_size, len(encodings.attrs2int))
        self._xpos = LinearNorm(64 + NUM_FILTERS // 2 + config.lang_emb_size, len(encodings.xpos2int))
        self._upos_emb = nn.Embedding(len(encodings.upos2int), 64)

        self._rnn = nn.LSTM(NUM_FILTERS // 2 + config.lang_emb_size + config.external_proj_size, config.rnn_size,
                            num_layers=config.rnn_layers, batch_first=True, bidirectional=True, dropout=0.33)

        self._pre_out = LinearNorm(config.rnn_size * 2 + config.lang_emb_size, config.pre_parser_size)
        # self._head_r1 = LinearNorm(config.pre_parser_size, config.head_size)
        # self._head_r2 = LinearNorm(config.pre_parser_size, config.head_size)
        # self._label_r1 = LinearNorm(config.pre_parser_size, config.label_size)
        # self._label_r2 = LinearNorm(config.pre_parser_size, config.label_size)
        # self._att_net = BilinearAttention(config.head_size, config.head_size)
        # self._label_linear = nn.Linear(config.label_size * 2, len(encodings.label2int))
        # self._label_bilinear = nn.Bilinear(config.label_size, config.label_size, len(encodings.label2int))
        self._head = DeepBiaffine(config.pre_parser_size, config.pre_parser_size, config.head_size, 1, dropout=0.1)
        self._label = DeepBiaffine(config.pre_parser_size, config.pre_parser_size, config.label_size,
                                   len(encodings.label2int), dropout=0.1, pairwise=False)
        self._r_emb = nn.Embedding(1,
                                   config.char_filter_size // 2 + config.lang_emb_size + config.word_emb_size + config.external_proj_size)

        # self._decoder = GreedyDecoder()
        self._decoder = ChuLiuEdmondsDecoder()

        if self._language_codes:
            self._res = {}
            for language_code in self._language_codes:
                self._res[language_code] = {"upos": 0., "attrs": 0., 'uas': 0., 'las': 0.}
            self._early_stop_meta_val = 0

    def _compute_early_stop(self, res):
        for lang in res:
            if res[lang]["uas"] > self._res[lang]["uas"]:
                self._early_stop_meta_val += 1
                self._res[lang]["uas"] = res[lang]["uas"]
                res[lang]["uas_best"] = True
            if res[lang]["las"] > self._res[lang]["las"]:
                self._early_stop_meta_val += 1
                self._res[lang]["las"] = res[lang]["las"]
                res[lang]["las_best"] = True
        return res

    def forward(self, X):
        x_sents = X['x_sent']
        x_lang_sent = X['x_lang_sent']
        x_words_chars = X['x_word']
        x_words_case = X['x_word_case']
        x_lang_word = X['x_lang_word']
        x_sent_len = X['x_sent_len']
        x_word_len = X['x_word_len']
        x_sent_masks = X['x_sent_masks']
        x_word_masks = X['x_word_masks']
        x_word_emb_packed = X['x_word_embeddings']
        gs_upos = None
        if 'y_upos' in X:
            gs_upos = X['y_upos']
        char_emb_packed = self._word_net(x_words_chars, x_words_case, x_lang_word, x_word_masks, x_word_len)

        sl = x_sent_len.cpu().numpy()

        char_emb = unpack(char_emb_packed, sl, x_sents.shape[1], device=self._get_device())
        word_emb_ext = None

        for ii in range(len(x_word_emb_packed)):
            we = unpack(x_word_emb_packed[ii], sl, x_sents.shape[1], self._get_device())
            if word_emb_ext is None:
                word_emb_ext = self._ext_proj[ii](we.float())
            else:
                word_emb_ext = word_emb_ext + self._ext_proj[ii](we.float())

        word_emb_ext = word_emb_ext / len(x_word_emb_packed)
        word_emb_ext = torch.tanh(word_emb_ext)

        lang_emb = self._lang_emb(x_lang_sent)
        lang_emb = lang_emb.unsqueeze(1).repeat(1, char_emb.shape[1] + 1, 1)

        aupos = self._aupos(torch.cat([char_emb, lang_emb[:, 1:, :]], dim=-1))
        aattrs = self._aattrs(torch.cat([char_emb, lang_emb[:, 1:, :]], dim=-1))
        axpos = self._axpos(torch.cat([char_emb, lang_emb[:, 1:, :]], dim=-1))

        word_emb = self._word_emb(x_sents)

        x = mask_concat([word_emb, char_emb, word_emb_ext], 0.33, self.training, self._get_device())

        x = torch.cat([x, lang_emb[:, 1:, :]], dim=-1)
        # prepend root
        root_emb = self._r_emb(torch.zeros((x.shape[0], 1), device=self._get_device(), dtype=torch.long))
        x = torch.cat([root_emb, x], dim=1)
        x = x.permute(0, 2, 1)
        lang_emb = lang_emb.permute(0, 2, 1)
        half = self._config.cnn_filter // 2
        res = None
        hidden = None
        cnt = 0
        for conv in self._convs:
            conv_out = conv(x)
            tmp = torch.tanh(conv_out[:, :half, :]) * torch.sigmoid((conv_out[:, half:, :]))
            if res is None:
                res = tmp
            else:
                res = res + tmp
            x = torch.dropout(tmp, 0.2, self.training)
            cnt += 1
            if cnt == self._config.aux_softmax_location:
                hidden = torch.cat([x + res, lang_emb], dim=1)
            if cnt != self._config.cnn_layers:
                x = torch.cat([x, lang_emb], dim=1)

        x = x + res
        x_parse = x.permute(0, 2, 1)
        # aux tagging
        lang_emb = lang_emb.permute(0, 2, 1)
        hidden = hidden.permute(0, 2, 1)[:, 1:, :]
        pre_morpho = torch.dropout(torch.tanh(self._pre_morpho(hidden)), 0.33, self.training)
        pre_morpho = torch.cat([pre_morpho, lang_emb[:, 1:, :]], dim=2)
        upos = self._upos(pre_morpho)
        if gs_upos is None:
            upos_idx = torch.argmax(upos, dim=-1)
        else:
            upos_idx = gs_upos

        upos_emb = self._upos_emb(upos_idx)
        attrs = self._attrs(torch.cat([pre_morpho, upos_emb], dim=-1))
        xpos = self._xpos(torch.cat([pre_morpho, upos_emb], dim=-1))

        # parsing
        word_emb_ext = torch.cat(
            [torch.zeros((word_emb_ext.shape[0], 1, self._config.external_proj_size), device=self._get_device(),
                         dtype=torch.float), word_emb_ext], dim=1)
        x = mask_concat([x_parse, word_emb_ext], 0.33, self.training, self._get_device())
        x = torch.cat([x, lang_emb], dim=-1)
        output, _ = self._rnn(x)
        output = torch.cat([output, lang_emb], dim=-1)
        pre_parsing = torch.dropout(torch.tanh(self._pre_out(output)), 0.33, self.training)
        # h_r1 = torch.tanh(self._head_r1(pre_parsing))
        # h_r2 = torch.tanh(self._head_r2(pre_parsing))
        # l_r1 = torch.tanh(self._label_r1(pre_parsing))
        # l_r2 = torch.tanh(self._label_r2(pre_parsing))
        # att_stack = []
        # for ii in range(1, h_r1.shape[1]):
        #     a = self._att_net(h_r1[:, ii, :], h_r2)
        #     att_stack.append(a.unsqueeze(1))
        # att = torch.cat(att_stack, dim=1)
        heads = self._head(pre_parsing, pre_parsing)
        labels = pre_parsing  # self._label(pre_parsing, pre_parsing)
        return heads.squeeze(-1)[:, 1:, :], labels, upos, xpos, attrs, aupos, axpos, aattrs

    def _get_labels(self, labels, heads):
        x1 = labels
        labs = []
        for ii in range(labels.shape[0]):
            lab = []
            for jj in range(1, labels.shape[1]):
                if jj <= len(heads[ii]):
                    lab.append(labels[ii, heads[ii][jj - 1]].unsqueeze(0).unsqueeze(0))
                else:
                    lab.append(labels[ii, 0].unsqueeze(0).unsqueeze(0))
            lab = torch.cat(lab, dim=1)
            labs.append(lab)
        x2 = torch.cat(labs, dim=0)

        labs = self._label(x1[:, 1:, :], x2)
        return labs

    # def _get_labels(self, x1, x2, heads):
    #     x1 = x1[:, 1:, :]
    #     x_stack = []
    #     for ii in range(x1.shape[0]):
    #         xx = []
    #         for jj in range(x1.shape[1]):
    #             if jj < len(heads[ii]):
    #                 xx.append(x2[ii, heads[ii][jj]].unsqueeze(0).unsqueeze(0))
    #             else:
    #                 xx.append(x2[ii, 0].unsqueeze(0).unsqueeze(0))
    #         x_stack.append(torch.cat(xx, dim=1))
    #     x_stack = torch.cat(x_stack, dim=0).contiguous()
    #     x1 = x1.contiguous()
    #     hid = torch.cat([x1, x_stack], dim=-1)
    #     return self._label_linear(hid) + self._label_bilinear(x1, x_stack)

    def _get_device(self):
        if self._lang_emb.weight.device.type == 'cpu':
            return 'cpu'
        return '{0}:{1}'.format(self._lang_emb.weight.device.type, str(self._lang_emb.weight.device.index))

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=1e-3, weight_decay=1e-4)
        return optimizer

    def training_step(self, batch, batch_idx):
        att, labels, p_upos, p_xpos, p_attrs, a_upos, a_xpos, a_attrs = self.forward(batch)
        y_upos = batch['y_upos']
        y_attrs = batch['y_attrs']
        y_xpos = batch['y_xpos']
        y_head = batch['y_head']
        y_label = batch['y_label']

        pred_labels = self._get_labels(labels, y_head.detach().cpu().numpy())

        loss_upos = F.cross_entropy(p_upos.view(-1, p_upos.shape[2]), y_upos.view(-1), ignore_index=0)
        loss_xpos = F.cross_entropy(p_xpos.view(-1, p_xpos.shape[2]), y_xpos.view(-1), ignore_index=0)
        loss_attrs = F.cross_entropy(p_attrs.view(-1, p_attrs.shape[2]), y_attrs.view(-1), ignore_index=0)

        loss_aupos = F.cross_entropy(a_upos.view(-1, a_upos.shape[2]), y_upos.view(-1), ignore_index=0)
        loss_axpos = F.cross_entropy(a_xpos.view(-1, a_xpos.shape[2]), y_xpos.view(-1), ignore_index=0)
        loss_aattrs = F.cross_entropy(a_attrs.view(-1, a_attrs.shape[2]), y_attrs.view(-1), ignore_index=0)

        loss_uas = F.cross_entropy(att.reshape(-1, att.shape[2]), y_head.view(-1))
        loss_las = F.cross_entropy(pred_labels.view(-1, pred_labels.shape[2]), y_label.view(-1), ignore_index=0)

        step_loss = loss_uas + loss_las + (((loss_upos + loss_attrs + loss_xpos) / 3.) + (
                (loss_aupos + loss_aattrs + loss_axpos) / 3.))

        return {'loss': step_loss}

    def validation_step(self, batch, batch_idx):
        y_upos = batch['y_upos']
        del batch['y_upos']
        att, labels, p_upos, p_xpos, p_attrs, a_upos, a_xpos, a_attrs = self.forward(batch)
        y_xpos = batch['y_xpos']
        y_attrs = batch['y_attrs']
        y_head = batch['y_head']
        y_label = batch['y_label']
        x_sent_len = batch['x_sent_len']
        x_lang = batch['x_lang_sent']
        sl = x_sent_len.detach().cpu().numpy()

        att = torch.softmax(att, dim=-1).detach().cpu().numpy()
        pred_heads = self._decoder.decode(att, sl)
        pred_labels = self._get_labels(labels, pred_heads)

        loss_upos = F.cross_entropy(p_upos.view(-1, p_upos.shape[2]), y_upos.view(-1), ignore_index=0)
        loss_attrs = F.cross_entropy(p_attrs.view(-1, p_attrs.shape[2]), y_attrs.view(-1), ignore_index=0)
        loss = (loss_upos + loss_attrs) / 3
        language_result = {lang_id: {'total': 0, 'upos_ok': 0, 'xpos_ok': 0, 'attrs_ok': 0, 'uas_ok': 0, 'las_ok': 0}
                           for lang_id in range(self._num_langs)}

        pred_upos = torch.argmax(p_upos, dim=-1).detach().cpu().numpy()
        pred_attrs = torch.argmax(p_attrs, dim=-1).detach().cpu().numpy()
        pred_xpos = torch.argmax(p_xpos, dim=-1).detach().cpu().numpy()
        pred_labels = torch.argmax(pred_labels, dim=-1).detach().cpu().numpy()
        tar_upos = y_upos.detach().cpu().numpy()
        tar_attrs = y_attrs.detach().cpu().numpy()
        tar_xpos = y_xpos.detach().cpu().numpy()
        tar_head = y_head.detach().cpu().numpy()
        tar_label = y_label.detach().cpu().numpy()

        x_lang = x_lang.detach().cpu().numpy()
        for iSent in range(p_upos.shape[0]):
            for iWord in range(sl[iSent]):
                lang_id = x_lang[iSent] - 1
                language_result[lang_id]['total'] += 1
                if pred_upos[iSent, iWord] == tar_upos[iSent, iWord]:
                    language_result[lang_id]['upos_ok'] += 1
                if pred_attrs[iSent, iWord] == tar_attrs[iSent, iWord]:
                    language_result[lang_id]['attrs_ok'] += 1
                if pred_xpos[iSent, iWord] == tar_xpos[iSent, iWord]:
                    language_result[lang_id]['xpos_ok'] += 1
                if pred_heads[iSent][iWord] == tar_head[iSent, iWord]:
                    language_result[lang_id]['uas_ok'] += 1
                    if pred_labels[iSent, iWord] == tar_label[iSent, iWord]:
                        language_result[lang_id]['las_ok'] += 1

        return {'loss': loss, 'acc': language_result}

    def validation_epoch_end(self, outputs):
        language_result = {lang_id: {'total': 0, 'upos_ok': 0, 'attrs_ok': 0, 'xpos_ok': 0, 'uas_ok': 0, 'las_ok': 0}
                           for lang_id in
                           range(self._num_langs)}

        valid_loss_total = 0
        total = 0
        attrs_ok = 0
        upos_ok = 0
        xpos_ok = 0
        uas_ok = 0
        las_ok = 0
        for out in outputs:
            valid_loss_total += out['loss']
            for lang_id in language_result:
                valid_loss_total += out['loss']
                language_result[lang_id]['total'] += out['acc'][lang_id]['total']
                language_result[lang_id]['upos_ok'] += out['acc'][lang_id]['upos_ok']
                language_result[lang_id]['xpos_ok'] += out['acc'][lang_id]['xpos_ok']
                language_result[lang_id]['attrs_ok'] += out['acc'][lang_id]['attrs_ok']
                language_result[lang_id]['uas_ok'] += out['acc'][lang_id]['uas_ok']
                language_result[lang_id]['las_ok'] += out['acc'][lang_id]['las_ok']
                # global
                total += out['acc'][lang_id]['total']
                upos_ok += out['acc'][lang_id]['upos_ok']
                xpos_ok += out['acc'][lang_id]['xpos_ok']
                attrs_ok += out['acc'][lang_id]['attrs_ok']
                uas_ok += out['acc'][lang_id]['uas_ok']
                las_ok += out['acc'][lang_id]['las_ok']

        self.log('val/loss', valid_loss_total / len(outputs))
        self.log('val/UPOS/total', upos_ok / total)
        self.log('val/ATTRS/total', attrs_ok / total)
        self.log('val/XPOS/total', xpos_ok / total)
        self.log('val/UAS/total', uas_ok / total)
        self.log('val/LAS/total', las_ok / total)

        res = {}
        for lang_index in language_result:
            total = language_result[lang_index]['total']
            if total == 0:
                total = 1
            if self._language_codes is None:
                lang = lang_index
            else:
                lang = self._language_codes[lang_index]
            res[lang] = {
                "upos": language_result[lang_index]['upos_ok'] / total,
                "xpos": language_result[lang_index]['xpos_ok'] / total,
                "attrs": language_result[lang_index]['attrs_ok'] / total,
                "uas": language_result[lang_index]['uas_ok'] / total,
                "las": language_result[lang_index]['las_ok'] / total
            }

            self.log('val/UPOS/{0}'.format(lang), language_result[lang_index]['upos_ok'] / total)
            self.log('val/XPOS/{0}'.format(lang), language_result[lang_index]['xpos_ok'] / total)
            self.log('val/ATTRS/{0}'.format(lang), language_result[lang_index]['attrs_ok'] / total)
            self.log('val/UAS/{0}'.format(lang), language_result[lang_index]['uas_ok'] / total)
            self.log('val/LAS/{0}'.format(lang), language_result[lang_index]['las_ok'] / total)

        # single value for early stopping
        self._epoch_results = self._compute_early_stop(res)
        self.log('val/early_meta', self._early_stop_meta_val)

    def load(self, model_path: str, device: str = 'cpu'):
        self.load_state_dict(torch.load(model_path, map_location='cpu')['state_dict'])
        self.to(device)

    def process(self, doc: Document, collate: MorphoCollate, batch_size: int = 4, num_workers: int = 4) -> Document:
        self.eval()
        dataset = MorphoDataset(doc)

        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate.collate_fn,
                                shuffle=False, num_workers=num_workers)
        index = 0
        with torch.no_grad():
            for batch in dataloader:

                del batch['y_upos']
                for key in batch:
                    if isinstance(batch[key], torch.Tensor):
                        batch[key] = batch[key].to(self._device)
                att, labels, p_upos, p_xpos, p_attrs, a_upos, a_xpos, a_attrs = self.forward(batch)

                x_sent_len = batch['x_sent_len']
                sl = x_sent_len.detach().cpu().numpy()

                batch_size = p_upos.size()[0]

                att = torch.softmax(att, dim=-1).detach().cpu().numpy()
                pred_heads = self._decoder.decode(att, sl)
                pred_labels = self._get_labels(labels, pred_heads)
                pred_labels = torch.argmax(pred_labels.detach(), dim=-1).cpu()
                p_upos = torch.argmax(p_upos, dim=-1).detach().cpu().numpy()
                p_xpos = torch.argmax(p_xpos, dim=-1).detach().cpu().numpy()
                p_attrs = torch.argmax(p_attrs, dim=-1).detach().cpu().numpy()
                for sentence_index in range(batch_size):  # for each sentence
                    # print(f"at index {index+sentence_index}, sentence {sentence_index} has {batch['x_sent_len'][sentence_index]} words.")
                    for word_index in range(batch["x_sent_len"][sentence_index]):
                        head = pred_heads[sentence_index][word_index]
                        label_id = pred_labels[sentence_index][word_index]
                        doc.sentences[index + sentence_index].words[word_index].head = head
                        doc.sentences[index + sentence_index].words[word_index].label = self._encodings.labels[label_id]
                        doc.sentences[index + sentence_index].words[word_index].upos = self._encodings.upos_list[
                            p_upos[sentence_index, word_index]]
                        doc.sentences[index + sentence_index].words[word_index].xpos = self._encodings.xpos_list[
                            p_xpos[sentence_index, word_index]]
                        doc.sentences[index + sentence_index].words[word_index].attrs = self._encodings.attrs_list[
                            p_attrs[sentence_index, word_index]]

                index += batch_size

        return doc

    class PrintAndSaveCallback(pl.callbacks.Callback):
        def __init__(self, store_prefix):
            super().__init__()
            self.store_prefix = store_prefix

        def on_validation_end(self, trainer, pl_module):
            metrics = trainer.callback_metrics
            epoch = trainer.current_epoch

            for lang in pl_module._epoch_results:
                res = pl_module._epoch_results[lang]
                if "uas_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".uas")
                if "las_best" in res:
                    trainer.save_checkpoint(self.store_prefix + "." + lang + ".las")

            trainer.save_checkpoint(self.store_prefix + ".last")

            s = "{0:30s}\tUAS\tLAS\tUPOS\tXPOS\tATTRS".format("Language")
            print("\n\n\t" + s)
            print("\t" + ("=" * (len(s) + 16)))
            for lang in pl_module._language_codes:
                uas = metrics["val/UAS/{0}".format(lang)]
                las = metrics["val/LAS/{0}".format(lang)]
                upos = metrics["val/UPOS/{0}".format(lang)]
                xpos = metrics["val/XPOS/{0}".format(lang)]
                attrs = metrics["val/ATTRS/{0}".format(lang)]
                msg = "\t{0:30s}:\t{1:.4f}\t{2:.4f}\t{3:.4f}\t{4:.4f}\t{5:.4f}".format(lang, uas, las, upos, xpos,
                                                                                       attrs)
                print(msg)
            print("\n")
